# file with my planet class
import os
import pandas as pd
import numpy as np

import astropy.units as u
from astropy import constants as const
from scipy.optimize import fsolve
import scipy.optimize as optimize

from platypos.lx_evo_and_flux import flux_at_planet, flux_at_planet_earth
from platypos.mass_evolution_function import mass_planet_RK4_forward_Ot20

class Planet_Ot20():
    """
    Structure of star_dictionary: {'star_id': "dummySun", 
                                   'mass': mass_star, 
                                   'radius': radius_star, 
                                   'age': age_star, 
                                   'L_bol': L_bol, 
                                   'Lx_age': Lx_age}
    Structure of planet_dict: {"radius": r_p, distance": a}
    """

    def __init__(self, star_dictionary, planet_dict, Lx_sat_info=None):
        '''Initialize a new instance (=object) of Planet_Ot20

        '''

        # initialize stellar parameters
        self.star_id = star_dictionary['star_id']
        self.mass_star = star_dictionary['mass']
        self.radius_star = star_dictionary['radius']
        self.age = star_dictionary['age']
        self.Lbol = star_dictionary['L_bol'] * const.L_sun.cgs.value
        self.Lx_age = star_dictionary['Lx_age']
        
        # initialize planet
        # -> based on observed radius, we estimate the mass from M-R relation
        # following params are the same for all planets!
        self.distance = planet_dict["distance"]
        self.flux = flux_at_planet_earth(self.Lbol, self.distance)
        self.has_evolved = False # flag which tracks if planet has been 
                                 # evolved and result file exists
        self.planet_id = "dummy" # set planet name later with set_name
        
        self.radius = planet_dict["radius"]
        self.mass_planet_Ot20()
        self.Lx_sat_info = Lx_sat_info

    def mass_planet_Ot20(self):
        """ calculate planet mass based on radius.
        I only use the volatile rich regime! This means the radius needs 
        to be greater than ~2.115 R_earth (radius with density > 3.3)
        """
        M_EARTH = const.M_earth.cgs.value
        R_EARTH = const.R_earth.cgs.value

        if (type(self.radius) == int) \
                or (type(self.radius) == float) \
                or (type(self.radius) == np.float64): # if R is single value
            M_p_volatile = 1.74 * (self.radius)**1.58 # if rho < 3.3 g/cm^3
            rho_volatile = M_p_volatile * M_EARTH \
                           / (4 / 3 * np.pi * (self.radius * R_EARTH)**3)
            if (rho_volatile >= 3.3):
                raise Exception("Planet with this radius is too small" \
                                + "and likely rocky; use LoFo14 models" \
                                + "instead.")
            else:
                if (M_p_volatile >= 120.):
                    raise Exception("Planet too massive. M-R relation only" \
                                    + "valid for <120 M_EARTH.")
                else:
                    self.mass = M_p_volatile

    def radius_planet_Ot20(self):
        """ calculate planet radius based on mass.
        I only use the volatile rich regime! This means the mass needs 
        to be bewteen ~5.7 and 120 M_earth
        """
        M_EARTH = const.M_earth.cgs.value
        R_EARTH = const.R_earth.cgs.value

        if (type(self.mass)==int) \
                or (type(self.mass) == float) \
                or (type(self.mass) == np.float64): # if M is single value
            R_p_volatile = (self.mass / 1.74)**(1. / 1.58)
            rho_volatile = self.mass * M_EARTH \
                           / (4 / 3 * np.pi * (R_p_volatile * R_EARTH)**3)
            if (rho_volatile >= 3.3):
                raise Exception("Planet with this mass/radius is too small" \
                                + "and likely rocky; use LoFo14 models" \
                                + "instead.")
            else:
                if (self.mass >= 120.):
                    raise Exception("Planet too massive. M-R relation only" \
                                    + "valid for <120 M_earth.")
                else:
                    self.radius = R_p_volatile

    def set_name(self, t_final, initial_step_size, 
                 epsilon, K_on, beta_on, evo_track_dict):
        """ function to set the right planet name based on the 
        track specified. This can then be used to check if a particular
        planet on a particular track has already evolved. 
        """
        self.planet_id = "planet_" \
                         + str(np.round(self.distance, 3)).replace('.', 'dot') \
                         + 'radius' \
                         + str(np.round(self.radius,3)).replace(".", "dot") \
                         + '_mass' + str(np.round(self.mass,3)) \
                         + '_Mstar' \
                         + str(np.round(self.mass_star,3)).replace(".", "dot") \
                         + "_K_" + K_on + "_beta_" + beta_on \
                         + "_track_" + str(evo_track_dict["t_start"]) + "_" \
                         + str(evo_track_dict["t_sat"]) + "_" \
                         + str(t_final) + "_" + str(evo_track_dict["Lx_max"]) \
                         + "_" + str(evo_track_dict["dt_drop"]) \
                         + "_" + str(evo_track_dict["Lx_drop_factor"])

    def generate_planet_id(self, t_final, 
                           planet_folder_id, evo_track_dict):
        """ Similar as set_name; produces a planet_id which can be used 
                for saving the results. """
        self.planet_id = planet_folder_id + "_track_" \
                         + str(evo_track_dict["t_start"]) + "_" \
                         + str(evo_track_dict["t_sat"]) + "_" \
                         + str(t_final) + "_" + str(evo_track_dict["Lx_max"]) \
                         + "_" + str(evo_track_dict["dt_drop"]) + "_" \
                         + str(evo_track_dict["Lx_drop_factor"])
    

    def write_general_params_to_file(self, path_for_saving, 
                                     planet_folder_id, evo_track_dict):
        """produces various files:
            - a file with the initial planet parameters (name is the
                planet folder name, e.g. planet_001.txt), 
            - a file with the host star parameters (host_star.txt), 
            - a file with the corresponding track parameters starting with 
                "track_params_"
        """ 
        # create a file (e.g. planet_XXXX.txt) which contains the
        # initial planet params
        if not os.path.exists(path_for_saving + planet_folder_id + ".txt"):
            with open(path_for_saving + planet_folder_id + ".txt", "w") as p:
                planet_params = "a,mass,radius,age\n" \
                                 + str(self.distance) + "," \
                                 + str(self.mass) + "," \
                                 + str(self.radius) + "," \
                                 + str(self.age)
                p.write(planet_params)

        # create a file (track_params_planet_....txt) which contains
        # the track parameters
        if not os.path.exists(path_for_saving + "track_params_" \
                              + self.planet_id + ".txt"):
                with open(path_for_saving + "track_params_" + self.planet_id + ".txt", "w") as t:
                    track_params = "t_start,t_sat,t_curr,t_5Gyr,Lx_max,Lx_curr," \
                                    + "Lx_5Gyr,dt_drop,Lx_drop_factor\n" \
                                    + str(evo_track_dict["t_start"]) + "," \
                                    + str(evo_track_dict["t_sat"]) + ","  \
                                    + str(evo_track_dict["t_curr"]) + "," \
                                    + str(evo_track_dict["t_5Gyr"]) + "," \
                                    + str(evo_track_dict["Lx_max"]) + "," \
                                    + str(evo_track_dict["Lx_curr"]) + "," \
                                    + str(evo_track_dict["Lx_5Gyr"]) + "," \
                                    + str(evo_track_dict["dt_drop"]) + "," \
                                    + str(evo_track_dict["Lx_drop_factor"])
                    t.write(track_params)

        # create a file which contains the host star parameters
        if not os.path.exists(path_for_saving + "host_star_properties.txt"):
            with open(path_for_saving + "host_star_properties.txt", "w") as s:
                star_params = "star_id,mass_star,radius_star,age,Lbol,Lx_age\n" \
                                + self.star_id + "," + str(self.mass_star) \
                                + "," + str(self.radius_star) \
                                + "," + str(self.age) \
                                + "," + str(self.Lbol / const.L_sun.cgs.value) \
                                + "," + str(self.Lx_age)
                s.write(star_params)

    def write_final_params_to_file(self, results_df, path_for_saving):
        """ Create file with only the final time, mass and radius parameters. """ 
        
        # create another file, which contains the final parameters only
        if not os.path.exists(path_for_saving \
                              + self.planet_id + "_final.txt"):
            with open(path_for_saving + self.planet_id + "_final.txt", "w") as p:
                # get last element (final time, mass, radius)
                t_final = results_df["Time"].iloc[-1]
                R_final = results_df["Radius"].iloc[-1]
                M_final = results_df["Mass"].iloc[-1]
                planet_params = "a,time,mass,radius,track\n" \
                                + str(self.distance) + "," \
                                + str(t_final) + "," \
                                + str(M_final) + "," \
                                + str(R_final) + "," \
                                + self.planet_id
                p.write(planet_params)

    def evolve_forward(self, t_final,
                       initial_step_size,
                       epsilon, K_on, beta_on,
                       evo_track_dict, 
                       path_for_saving,
                       planet_folder_id):
        """ Call this function to make the planet evolve and 
        create file with mass and radius evolution. 
        See Mass_evolution_function.py for details on the integration.
        """
        
        if os.path.exists(path_for_saving + self.planet_id + ".txt"):
            # planet already exists
            self.has_evolved = True
            df = pd.read_csv(path_for_saving + self.planet_id + ".txt")
        else:
            #print("Planet: ", self.planet_id+".txt")
            # call mass_planet_RK4_forward_LO14 to start the integration
            t, M, R, Lx = mass_planet_RK4_forward_Ot20(
                                    epsilon=epsilon,
                                    K_on=K_on,
                                    beta_on=beta_on,
                                    planet_object=self,
                                    initial_step_size=initial_step_size,
                                    t_final=t_final,
                                    track_dict=evo_track_dict
                                    )
            
            ### TO DO: move this to mass_planet_RK4_forward_LO14 -
            # > make it return dataframe
            # add results to dataframe and save
            df = pd.DataFrame({"Time": t, "Mass": M, "Radius": R, "Lx": Lx})
            df.to_csv(path_for_saving + self.planet_id + ".txt", index=None)
            self.has_evolved = True  # set evolved-flag to True
                
        return df
    
    def evolve_forward_and_create_full_output(self, t_final,
                                              initial_step_size,
                                              epsilon,
                                              K_on,
                                              beta_on,
                                              evo_track_dict,
                                              path_for_saving,
                                              planet_folder_id
                                              ):
        """ This is the master-function which needs to be called to 
        evolve the planet and at the same time create all necessary 
        output files which contain useful data about initial params, 
        host star params, tracks etc..."""
        
        # create planet id for file names
        self.generate_planet_id(t_final, planet_folder_id, evo_track_dict)
        # create file with initial planet, host star and track params
        self.write_general_params_to_file(path_for_saving,
                                          planet_folder_id,
                                          evo_track_dict)
        
        results_df = self.evolve_forward(t_final, initial_step_size,
                                         epsilon, K_on, beta_on,
                                         evo_track_dict,
                                         path_for_saving,
                                         planet_folder_id)
        
        # create file with final planet params
        self.write_final_params_to_file(results_df, path_for_saving)

    def read_results(self, file_path):
        """  read in results file and return dataframe. """
        # planet exists, has been evolved and results can be read in
        if self.has_evolved == True:
            df = pd.read_csv(file_path+self.planet_id+".txt",
                                             float_precision='round_trip')
            # 'round_trip': to avoid pandas doing sth. weird to the last digit
            # Info: Pandas uses a dedicated dec 2 bin converter that
            # a compromisesccuracy in preference to speed.
            # Passing float_precision='round_trip' to read_csv fixes this.
            #t, M, R, Lx = df["Time"].values, df["Mass"].values,\
            #              df["Radius"].values, df["Lx"].values
            return df #t, M, R, Lx
        else:
            print("Planet has not been evolved & no result file exists.")
                        
    def evolve_backwards(self):
            raise NotImplementedError("Coming soon! :)")



# Lopez & Fortney 2014 models
import numpy as np
import warnings
import astropy.units as u
from astropy import constants as const

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
# ignoring this warining is certainly not the best way, but it is currently
# thrown when one of the "in-between" time steps inside the Runge-Kutta
# integration method results in a planet which has no atmosphere left.


def calculate_core_radius(M_core):
    """ M-R relation for rock/iron Earth-like core. (no envelope)
    (see Lopez & Fortney 2014 for details.)
    """
    R_core = (M_core**0.25)
    return R_core


def calculate_planet_mass(M_core, fenv):
    """ Planet mass determined by core mass and atmosphere mass
    (specified in terms of envelope mass fraction [in % !]). """
    M_pl = M_core/(1-(fenv/100))
    return M_pl


def calculate_R_env(M_p, fenv, F_p, age, metallicity):
    """ Calculate planetary envelope radius using the parametrized
    results from Lopez & Fortney 2014 for planets with rocky core
    and H/He gaseous envelope.

    Parameters:
    -----------
    M_p (float): planet mass (in Earth masses)
    f_env (float): envelope mass fraction (given in percent!)
    F_p (float): bolometric flux revcieved by the planet (in units
                 of Earth's insolation)
    age (float): age of the planet (in Myr)
    metallicity (str): chose models with solar or enhanced metallicity
                       (set to "solarZ" or "enhZ")
    (check Lopez & Fortney 2014 for further details and range of model
    validity).

    Returns:
    --------
    R_env (float): radius of the envelope (in Earth radii)
    """
    age_exponent = {"solarZ": -0.11, "enhZ": -0.18}
    R_env = 2.06 * (M_p)**(-0.21) * (fenv/5)**0.59 * (F_p)**0.044 \
        * ((age/1e3)/5)**(age_exponent[metallicity])
    return R_env  # R_earth


def calculate_planet_radius(M_core, fenv, age,
                            flux_solar, metallicity):
    """ Calculate planetary radius (core + envelope) using the
    parametrized results from Lopez & Fortney 2014 for planets with
    rocky core and H/He gaseous envelope.

    Parameters:
    -----------
    M_core (float): core mass (in Earth masses)
    f_env (float): envelope mass fraction (given in percent!)
    flux_solar (float): bolometric flux revcieved by the planet
                        (in units of Earth's flux)
    age (float): age of the planet (in Myr)
    metallicity (str): chose models with solar or enhanced metallicity
                       (set to "solarZ" or "enhZ")
    (check Lopez & Fortney 2014 for further details and range of model
    validity).

    Returns:
    --------
    R_pl (float): radius of the planet (in Earth radii)
    """

    R_core = calculate_core_radius(M_core)
    M_pl = calculate_planet_mass(M_core, fenv)
    R_env = calculate_R_env(M_pl, fenv, flux_solar, age, metallicity)
    R_pl = R_core + R_env
    return R_pl


def density_planet(M_p, R_p):
    """ Calculate the mean density in cgs units using the radius and
    mass of the planet.

    Parameters:
    -----------
    M_p (float): mass of the planet (in Earth masses)
    R_p (float): radius of the planet (in Earth radii)

    Returns:
    --------
    rho (float): density in cgs units (g/ccm)
    """

    rho = (M_p*const.M_earth.cgs) \
        / (4./3*np.pi*(R_p*const.R_earth.cgs)**3)
    return rho.value



import numpy as np
import astropy.units as u
from astropy import constants as const


def beta_fct(M_p, F_xuv, R_p):
    """Function estimates the beta parameter, which is a correction to the
    planetary absorption radius in XUV, as planet appers larger than in
    optical; this approximation comes from a study by Salz et al. (2016)
    NOTE from paper: "The atmospheric expansion can be neglected for massive
    hot Jupiters, but in the range of superEarth-sized planets the expansion
    causes mass-loss rates that are higher by a factor of four."

    Parameters:
    ----------
    M_p (float): mass of the planet in Earth units
    F_xuv (float or list): XUV flux recieved by planet
    R_p (float): radius mass of the planet in Earth units

    Returns:
    --------
    beta (float or array): beta parameter
    """

    M_EARTH= const.M_earth.cgs.value
    R_EARTH = const.R_earth.cgs.value

    if (type(F_xuv) == float) or (type(F_xuv) == np.float64):
        # if F_xuv is single value
        grav_pot = -const.G.cgs.value * (M_p*M_EARTH) / (R_p*R_EARTH)
        log_beta = max(0.0, -0.185 * np.log10(-grav_pot)
        					+ 0.021 * np.log10(F_xuv) + 2.42)
        beta = 10**log_beta
        return beta

    elif len(F_xuv) > 1:
        # if F_xuv is a list
        betas = []
        for i in range(len(F_xuv)):
            grav_pot_i = -const.G.cgs.value \
                         * (M_p[i]*M_EARTH) / (R_p[i]*R_EARTH)
            log_beta_i = max(0.0, -0.185 * np.log10(-grav_pot_i)
            					  + 0.021 * np.log10(F_xuv[i]) + 2.42)
            beta_i = 10**log_beta_i
            betas.append(beta_i)
        betas = np.array(betas)
        return betas


def K_fct(a_pl, M_pl, M_star, R_pl):
    """ K correction factor to take into account tidal influences of the
    host star on the planetary atmosphere (from Erkaev et al. 2007).

    Parameters:
    -----------
    a_pl (float): star-planet separation in A.U.
    M_pl (float or list): mass of the planet in Earth units
    M_star (float): mass of the star in solar units
    R_pl (float): radius of the planet in Earth units

    Returns:
    --------
    K (float) or K (array)
    """

    AU = const.au.cgs.value
    M_SUN = const.M_sun.cgs.value
    M_EARTH = const.M_earth.cgs.value
    R_EARTH = const.R_earth.cgs.value

    if (type(M_pl) == float) or (type(M_pl) == np.float64):
        # if M_pl is single value
        R_roche = (a_pl*AU) \
        		  * ((M_pl*M_EARTH) / (3*(M_star*M_SUN)))**(1./3)
        K = 1. - 3. * (R_pl*R_EARTH) / (2.*R_roche) \
			+ (R_pl*R_EARTH)**3. / (2. * R_roche**3.)
        return K

    elif len(M_pl) > 1:
        # if F_xuv is array
        Ks = []
        for i in range(len(M_pl)):
            R_roche_i = (a_pl*AU) \
            			* ((M_pl[i]*M_EARTH) / (3*(M_star*M_SUN)))**(1./3)
            K_i = 1. - 3. * (R_pl[i]*R_EARTH) / (2.*R_roche_i) \
                  + (R_pl[i]*R_EARTH)**3. / (2. * R_roche_i**3.)
            Ks.append(K_i)
        Ks = np.array(Ks)
        return Ks



# file with my planet class
import os
import numpy as np
import pandas as pd

import astropy.units as u
from astropy import constants as const
import scipy.optimize as optimize
from scipy.optimize import fsolve

from platypos.lx_evo_and_flux import flux_at_planet, flux_at_planet_earth
from platypos.mass_evolution_function import mass_planet_RK4_forward_LO14_PAPER


class Planet_LoFo14_PAPER():
    """
    Need star and planet dictionary to initialize a planet object.
    Structure of star_dictionary: {'star_id': "dummySun", 'mass': mass_star, 'radius': radius_star, 
                                   'age': age_star, 'L_bol': L_bol, 'Lx_age': Lx_age}
    Structure of planet_dict: {"core_mass": m_c, "fenv": f, "distance": a, "metallicity": metal}
    """
    def __init__(self, star_dictionary, planet_dict, Lx_sat_info=None):
        
        # initialize stellar parameters
        self.star_id = star_dictionary['star_id']
        self.mass_star = star_dictionary['mass']
        self.radius_star = star_dictionary['radius']
        self.age = star_dictionary['age']
        self.Lbol = star_dictionary['L_bol'] * const.L_sun.cgs.value
        self.Lx_age = star_dictionary['Lx_age']
        
        # initialize planet parameters
        # I have three options for planet_dict:
                # 1) artificial planet with M_core & f_env -> need to calculate M_pl & R_pl
                # 2) observed planet with a mass: M_pl & R_pl & f_env -> M_core 
                # 3) observed planet w/o mass: M_core, R_pl -> calulate f_env & M_pl
        
        # following 3 params are the same for all planets!
        self.distance = planet_dict["distance"]
        self.metallicity = planet_dict["metallicity"]  # solarZ or enhZ
        self.flux = flux_at_planet_earth(self.Lbol, self.distance)
        self.has_evolved = False # flag which tracks if planet has been evolved and result file exists
        self.planet_id = "None"
        self.Lx_sat_info = Lx_sat_info
        
        # the remaining parameters depend on if you choose a case 1, 2, or 3 planet
        while True:
            try:
                planet_dict["fenv"]
                #print("Case 1 - artificial planet (with M_core & f_env)") # Case 1
                self.planet_info = "Case 1 - artificial planet (with M_core & f_env)"
                # Case 1: artificial planet with fenv & M_core given -> need to calculate the planet mass
                self.fenv = planet_dict["fenv"]
                self.core_mass = planet_dict["core_mass"]
                self.calculate_core_radius()
                self.calculate_planet_mass()
                self.calculate_planet_radius()
                break
            except KeyError: # Case 2 or 3
                #print("No fenv given.")
                while True:
                    try:
                        planet_dict["mass"]
                        # Case 3 - observed planet with radius but no mass & M_core specified -> need to calculate/estimate fenv & M_pl
                        #print("Case 3 - obs. planet with radius & mass measurement (with R_pl, M_pl, M_core)")
                        self.planet_info = "Case 3 - obs. planet with radius & mass measurement (with R_pl, M_pl, M_core)"
                        self.core_mass = planet_dict["core_mass"]
                        self.mass = planet_dict["mass"]
                        self.radius = planet_dict["radius"]
                        self.calculate_core_radius()
                        self.solve_for_fenv() # function to solve for the envelope mass fraction
                        # add sanity check to make sure the mass with the calculated fenv matches the observed input mass!
                        # add sanity check to make sure the radius with the calculated fenv matches the observed input radius!
                        
                        break
                    except KeyError:
                        #print("Case 2 - obs. planet with radius, but no mass (with R_pl, M_core)") 
                        self.planet_info = "Case 2 - obs. planet with radius, but no mass (with R_pl, M_core)"
                        # Case 2 - observed planet with a mass, radius & core mass specified -> need to calculate fenv
                        self.core_mass = planet_dict["core_mass"]
                        self.radius = planet_dict["radius"]
                        self.calculate_core_radius()
                        self.solve_for_fenv() # function to solve for the envelope mass fraction
                        self.calculate_planet_mass()
                        # add sanity check to make sure the radius with the calculated fenv matches the observed input radius!
                        
                        break
                break

    # Class Methods
    def calculate_planet_mass(self):
        """ Planet mass determined by core mass and atmosphere mass (specified in terms of atm. mass fraction [%]). """
        #M_core = (self.core_mass/const.M_earth.cgs).value
        self.mass = self.core_mass/(1-(self.fenv/100))
    
    def calculate_planet_core_mass(self):
        """ Planet core mass determined by planet mass and atmosphere mass (specified in terms of atm. mass fraction [%]). """
        self.core_mass = self.mass*(1-(self.fenv/100))
     
    def calculate_core_radius(self):
        """M-R relation for rock/iron Earth-like core. (no envelope)"""
        self.core_radius = (self.core_mass**0.25)

    def solve_for_fenv(self):
        if self.radius == self.core_radius:
            self.fenv = 0.0
        else:
            def calculate_fenv(fenv):
                age_exponent = {"solarZ": -0.11, "enhZ": -0.18}  
                return -self.radius + self.core_radius + (2.06 * (self.core_mass/(1-(fenv/100)))**(-0.21) \
                                                        * (fenv/5)**0.59 * (self.flux)**0.044 * \
                                                          ((self.age/1e3)/5)**(age_exponent[self.metallicity]))
            f_guess = 0.1
            fenv = optimize.fsolve(calculate_fenv, x0=f_guess)[0]
            #print("fenv = ", fenv)
            self.fenv = fenv
            #if fenv1 != fenv:
            #    print("Sth went wrong in solving for the envelope mass fraction! Check!")

        
    def calculate_R_env(self):
        """ M_p in units of Earth masses, f_env in percent, Flux in earth units, age in Gyr
        R_env ~ t**0.18 for *enhanced opacities*;
        R_env ~ t**0.11 for *solar metallicities* 
        """
        age_exponent = {"solarZ": -0.11, "enhZ": -0.18}   
        R_env = 2.06 * self.mass**(-0.21) * (self.fenv/5)**0.59 * \
                self.flux**0.044 * ((self.age/1e3)/5)**(age_exponent[self.metallicity]) 
        return R_env # in units of R_earth


    def calculate_planet_radius(self):
        """ description: """
        self.radius = self.core_radius + self.calculate_R_env()
    
    def set_name(self, t_final, initial_step_size, 
                 epsilon, K_on, beta_on, evo_track_dict):
        """ function to set the right planet name based on the 
        track specified. This can then be used to check if a particular
        planet on a particular track has already evolved. 
        """
        self.planet_id = "planet_" \
                         + str(np.round(self.distance, 3)).replace('.', 'dot') \
                         + 'radius' \
                         + str(np.round(self.radius,3)).replace(".", "dot") \
                         + '_mass' + str(np.round(self.mass,3)) \
                         + '_Mstar' \
                         + str(np.round(self.mass_star,3)).replace(".", "dot") \
                         + "_K_" + K_on + "_beta_" + beta_on \
                         + "_track_" + str(evo_track_dict["t_start"]) + "_" \
                         + str(evo_track_dict["t_sat"]) + "_" \
                         + str(t_final) + "_" + str(evo_track_dict["Lx_max"]) \
                         + "_" + str(evo_track_dict["dt_drop"]) \
                         + "_" + str(evo_track_dict["Lx_drop_factor"])

    def generate_planet_id(self, t_final, 
                           planet_folder_id, evo_track_dict):
        """ Similar as set_name; produces a planet_id which can be used 
            for saving the results. """
        self.planet_id = planet_folder_id + "_track_" \
                         + str(evo_track_dict["t_start"]) + "_" \
                         + str(evo_track_dict["t_sat"]) + "_" \
                         + str(t_final) + "_" + str(evo_track_dict["Lx_max"]) \
                         + "_" + str(evo_track_dict["dt_drop"]) + "_" \
                         + str(evo_track_dict["Lx_drop_factor"])
    

    def write_general_params_to_file(self, path_for_saving, 
                                     planet_folder_id, evo_track_dict):
        """produces various files:
            - a file with the initial planet parameters (name is the
              planet folder name, e.g. planet_001.txt), 
            - a file with the host star parameters (host_star.txt), 
            - a file with the corresponding track parameters starting with 
              "track_params_"
        """ 
        # create a file (e.g. planet_XXXX.txt) which contains the
        # initial planet params
        if not os.path.exists(path_for_saving + planet_folder_id + ".txt"):
            with open(path_for_saving + planet_folder_id + ".txt", "w") as p:
                planet_params = "a,mass,radius,age\n" \
                                 + str(self.distance) + "," \
                                 + str(self.mass) + "," \
                                 + str(self.radius) + "," \
                                 + str(self.age)
                p.write(planet_params)

        # create a file (track_params_planet_....txt) which contains
        # the track parameters
        if not os.path.exists(path_for_saving + "track_params_" \
                              + self.planet_id + ".txt"):
            with open(path_for_saving + "track_params_" + self.planet_id + ".txt", "w") as t:
                track_params = "t_start,t_sat,t_curr,t_5Gyr,Lx_max,Lx_curr," \
                                + "Lx_5Gyr,dt_drop,Lx_drop_factor\n" \
                                + str(evo_track_dict["t_start"]) + "," \
                                + str(evo_track_dict["t_sat"]) + ","  \
                                + str(evo_track_dict["t_curr"]) + "," \
                                + str(evo_track_dict["t_5Gyr"]) + "," \
                                + str(evo_track_dict["Lx_max"]) + "," \
                                + str(evo_track_dict["Lx_curr"]) + "," \
                                + str(evo_track_dict["Lx_5Gyr"]) + "," \
                                + str(evo_track_dict["dt_drop"]) + "," \
                                + str(evo_track_dict["Lx_drop_factor"])
                t.write(track_params)

        # create a file which contains the host star parameters
        if not os.path.exists(path_for_saving + "host_star_properties.txt"):
            with open(path_for_saving + "host_star_properties.txt", "w") as s:
                star_params = "star_id,mass_star,radius_star,age,Lbol,Lx_age\n" \
                              + self.star_id + "," + str(self.mass_star) \
                              + "," + str(self.radius_star) \
                              + "," + str(self.age) \
                              + "," + str(self.Lbol / const.L_sun.cgs.value) \
                              + "," + str(self.Lx_age)
                s.write(star_params)

    def write_final_params_to_file(self, results_df, path_for_saving):
        """ Create file with only the final time, mass and radius parameters. """ 
        
        # create another file, which contains the final parameters only
        if not os.path.exists(path_for_saving \
                              + self.planet_id + "_final.txt"):
            with open(path_for_saving + self.planet_id + "_final.txt", "w") as p:
                # get last element (final time, mass, radius)
                t_final = results_df["Time"].iloc[-1]
                R_final = results_df["Radius"].iloc[-1]
                M_final = results_df["Mass"].iloc[-1]
                planet_params = "a,time,mass,radius,track\n" \
                                + str(self.distance) + "," \
                                + str(t_final) + "," \
                                + str(M_final) + "," \
                                + str(R_final) + "," \
                                + self.planet_id
                p.write(planet_params)

    def evolve_forward(self, t_final,
                       initial_step_size,
                       epsilon, K_on, beta_on,
                       evo_track_dict, 
                       path_for_saving,
                       planet_folder_id):
        """ Call this function to make the planet evolve and 
        create file with mass and radius evolution. 
        See Mass_evolution_function.py for details on the integration.
        """
        
        if os.path.exists(path_for_saving + self.planet_id + ".txt"):
            # planet already exists
            self.has_evolved = True
            df = pd.read_csv(path_for_saving + self.planet_id + ".txt")
        else:
            #print("Planet: ", self.planet_id+".txt")
            # call mass_planet_RK4_forward_LO14 to start the integration
            t, M, R, Lx = mass_planet_RK4_forward_LO14_PAPER(
                                    epsilon=epsilon,
                                    K_on=K_on,
                                    beta_on=beta_on,
                                    planet_object=self,
                                    initial_step_size=initial_step_size,
                                    t_final=t_final,
                                    track_dict=evo_track_dict
                                    )
            
            ### TO DO: move this to mass_planet_RK4_forward_LO14 -
            # > make it return dataframe
            # add results to dataframe and save
            df = pd.DataFrame({"Time": t, "Mass": M, "Radius": R, "Lx": Lx})
            df.to_csv(path_for_saving + self.planet_id + ".txt", index=None)
            self.has_evolved = True  # set evolved-flag to True
            
        return df

    def evolve_forward_and_create_full_output(self, t_final,
                                              initial_step_size,
                                              epsilon,
                                              K_on,
                                              beta_on,
                                              evo_track_dict,
                                              path_for_saving,
                                              planet_folder_id
                                              ):
        """ This is the master-function which needs to be called to 
        evolve the planet and at the same time create all necessary 
        output files which contain useful data about initial params, 
        host star params, tracks etc..."""
        
        # create planet id for file names
        self.generate_planet_id(t_final, planet_folder_id, evo_track_dict)
        # create file with initial planet, host star and track params
        self.write_general_params_to_file(path_for_saving,
                                          planet_folder_id,
                                          evo_track_dict)
        
        results_df = self.evolve_forward(t_final, initial_step_size,
                                         epsilon, K_on, beta_on,
                                         evo_track_dict,
                                         path_for_saving,
                                         planet_folder_id)
        
        # create file with final planet params
        self.write_final_params_to_file(results_df, path_for_saving)
        
    def read_results(self, file_path):
        """  read in results file and return dataframe. """
        # planet exists, has been evolved and results can be read in
        if self.has_evolved == True:
            df = pd.read_csv(file_path+self.planet_id+".txt",
                             float_precision='round_trip')
            # 'round_trip': to avoid pandas doing sth. weird to the last digit
            # Info: Pandas uses a dedicated dec 2 bin converter that
            # a compromisesccuracy in preference to speed.
            # Passing float_precision='round_trip' to read_csv fixes this.
            #t, M, R, Lx = df["Time"].values, df["Mass"].values,\
            #              df["Radius"].values, df["Lx"].values
            return df #t, M, R, Lx
        else:
            print("Planet has not been evolved & no result file exists.")
              
    def evolve_backwards(self):
        raise NotImplementedError("Coming soon! :)")


import numpy as np
import astropy.units as u
from astropy import constants as const
import scipy.optimize as optimize
from scipy.optimize import fsolve


def lx_evo(t, track_dict):
    """
    Function to calculate the X-ray luminosity, Lx, of the host star 
    at a given time t.
    NOTE: this function is not very pretty and could use some
    	  cosmetic updates.
    
    NOTE: I tried to set up this function such that it works going 
    forward (for young system) and backwards (for older system) in time;
    -> so t_curr(ent) is a little confusing! -> we have two cases!!
    
    Parameters Case 1 - a young system which you want to evolve forward in time 
    (i.e. estimate the mass loss the planet will undergo in the future):
    ---------------------------------------------------------------------------
    t_start (float): Starting time for the forward simulation in Myr 
    				 (e.g. 23 Myr for V1298 Tau)
    
    Lx_max (float): Measured Lx value at the system age (i.e. t_start)
    
    t_curr (float): Set to 1 Gyr (with Lx_curr = Lx at 1 Gyr), which is 
    				where, by design, all activity tracks converge
    				(based on Tu et al. (2015)/ Johnstone et al.(2015) models);

    Lx_curr (float): Lx value at 1 Gyr taken from Tu et al. (2015)
    
    t_5Gyr (float): Defines, together with Lx_5Gyr, the slope common for all 
    				activity evolution tracks past 1 Gyr 
    				(also based on Tu et al.(2015));
    
    Lx_5Gyr (float): Defines, together with t_5Gyr, the slope common for all 
    				 activity evolution tracks past 1 Gyr 
    				 (also based on Tu et al.(2015));
    Note: in principle these two pairs (t_curr, Lx_1Gyr) and (t_5Gyr, Lx_5Gyr)
    can be set however the user wants.
    
    # Case 2 not implemented yet in Platypos
    Parameters Case 2 - an older system, which you want to evolve backwards in time 
    (i.e. estimate the mass loss the planet has undergone until now):
    ---------------------------------------------------------------------------
    t_start is the youngest age you want to calculate backwards 
    (i.e. this shouldbe sth. like ~10 Myr, or close to the disk clearing 
    timescale); 

    Lx_max (float): Some saturation X-ray luminosity 
    				(e.g. Lx,sat/Lbol ~ 10^(-3))

    t_curr (float): Needs to be set to the system age 
    				(e.g. t_curr=700 Myr for K2-136b); 

    Lx_curr (float): If a measured Lx value is available for the system, 
    				 otherwise need to estimate;
    t_5Gyr (float): Defines, together with Lx_5Gyr, the slope common for all
        			activity evolution tracks past 1 Gyr 
        			(also based on Tu et al.(2015));
    Lx_5Gyr (float): Defines, together with t_5Gyr, the slope common for all
        			 activity evolution tracks past 1 Gyr 
        			 (also based on Tu et al.(2015));

    
    Returns:
    --------
    Lx_at_time_t (float) Returns the corresponding Lx value at given time t
    """
    
    # read in the parameters from the provided dictionary
    # -> these are all the parameters required to define the tracks
    t_start, t_sat=  track_dict["t_start"], track_dict["t_sat"]
    t_curr, t_5Gyr = track_dict["t_curr"], track_dict["t_5Gyr"]
    Lx_max = track_dict["Lx_max"]
    Lx_curr, Lx_5Gyr = track_dict["Lx_curr"], track_dict["Lx_5Gyr"]
    dt_drop = track_dict["dt_drop"]
    Lx_drop_factor = track_dict["Lx_drop_factor"]
    
    # Define function for calculating a power law
    powerlaw = lambda x, k, index: k * (x**index)
    
    # t_curr has to be greater than t_start
    if (t_curr < t_start) or (t < t_start):
        if t <= t_start: # produce a saturation regime before t_start
            Lx = Lx_max
        else:
            print("Make sure t_curr > t_start, t >= t_start")
    
    elif t > t_curr:
        # this is the regime past 1 Gyr (the same for all Tu tracks)
        # I use the Tu 2015 slope which is approx the same for all 
        # three evolutionary tracks (and based on current solar value);
        # this is mainly for consistency since we approximate our tracks
        # based on the Tu-tracks
        
        # exception for the OwWu17 case, where the slope past t_sat is c
        # onstant for all t > t_sat
        if t_5Gyr == t_curr: # then we are dealing with the OwWu17 case
            alpha = (np.log10(Lx_curr/Lx_max))/(np.log10(t_curr/t_sat))
            k = 10**(np.log10(Lx_max) - alpha*np.log10(t_sat))
            Lx = powerlaw(t, k, alpha)

        else: # normal case, i.e. slope past t_curr given by input parameters
            alpha = (np.log10(Lx_5Gyr/Lx_curr))/(np.log10(t_5Gyr/t_curr))
            k = 10**(np.log10(Lx_curr) - alpha*np.log10(t_curr))
            Lx = powerlaw(t, k, alpha)
        
        return Lx
    
    elif t_curr > t_start:
        # dt_drop==0 means we create a track with only the saturation regime
        # and a single power-law-slope drop to the converging Lx at t_curr
        if dt_drop==0: # then t_sat == t_drop
            if t_start >= t_sat: # then t_sat <= t_start < t_curr
                if t >= t_sat: #and t <= t_curr:
                    # only data in power law regime
                    alpha = (np.log10(Lx_curr / Lx_max)) \
                    		 / ( np.log10(t_curr / t_sat))
                    k = 10**(np.log10(Lx_max) - alpha * np.log10(t_sat))
                    Lx = powerlaw(t, k, alpha)
                    
            elif t_start < t_sat: # then t_start < t_sat < t_curr
                if t > t_sat: #and t <= t_curr:
                    alpha = (np.log10(Lx_curr / Lx_max)) \
                    		 / (np.log10(t_curr / t_sat))
                    k = 10**(np.log10(Lx_max) - alpha * np.log10(t_sat))
                    Lx = powerlaw(t, k, alpha)
                elif t <= t_sat:
                    Lx = Lx_max
                    
        elif dt_drop > 0: # then t_sat != t_drop
            # dt_drop > 0 means we create a more fancy track with the 
            # saturation regime and then a drop with slope change to 
            # the converging Lx at 1 Gyr (Lx track with three regimes)
            t_drop = t_sat + dt_drop # t_sat < t_drop
            
            if t <= t_sat:
                Lx = Lx_max
            elif (t > t_sat) and (t <= t_drop): # first of the two slopes
                alpha_drop1 = (np.log10((Lx_max / Lx_drop_factor) / Lx_max)) \
                			  / (np.log10(t_drop / t_sat))
                k_drop1 = 10**(np.log10(Lx_max) - alpha_drop1 \
                			   * np.log10(t_sat))
                Lx = powerlaw(t, k_drop1, alpha_drop1)
                
            elif t > t_drop:
                alpha_drop2 = (np.log10(Lx_curr / (Lx_max / Lx_drop_factor))) \
                			   / (np.log10(t_curr / t_drop))
                k_drop2 = 10**(np.log10((Lx_max / Lx_drop_factor)) \
                			   - alpha_drop2 * np.log10(t_drop))
                Lx = powerlaw(t, k_drop2, alpha_drop2)     
    return Lx

def flux_at_planet_earth(L, a_p):
    """Function calculates the flux that the planet recieves at a given 
    distance normalized with Earth's flux -> the semi-major axis is used
    for the distance (the eccentricity of the planetary orbits is ignored)
    """
    FLUX_AT_EARTH = 1373. * 1e7 * (u.erg/u.s) / (100*u.cm*100*u.cm) # erg/s/cm^2

    A = (4. * np.pi * (a_p * const.au.cgs)**2)
    F = (L * u.erg / u.s) / A
    F_earth = (F/FLUX_AT_EARTH).value
    return F_earth

def flux_at_planet(L, a_p):
    """Function calculates the flux that the planet recieves at a given 
    distance (in erg/s/cm^2) -> the semi-major axis is used for the distance
    (the eccentricity of the planetary orbits is ignored)
    """
    A = (4. * np.pi * (a_p * const.au.cgs)**2)
    F = ((L * u.erg / u.s) / A).value
    return F # erg/s/cm^2

def l_xuv_all(Lx):
    """Function to estimate the EUV luminosity using the scaling relations 
    given by Sanz-Forcada et al. (2011); the measured X-ray luminosity can 
    be extrapolated to the total high-energy radiation as given below. 
    To get the combined high-energy luminosity (L_xuv), we add L_x and L_euv.
    THEY USE: Lx(0.1-2.4 keV) ROSAT band
    """
    log_L_euv = 0.860 * np.log10(Lx) + 4.80
    Leuv = 10.**log_L_euv
    log_L_xuv = np.log10(Lx + Leuv)
    return 10.**log_L_xuv  # erg/s

def undo_what_Lxuv_all_does(L_xuv):
    """ If you have L_XUV given, this takes the Sanz-Forcada et al. (2011) 
    scaling relation and reverses it to estimate Lx.
    Function needs better implementation!
    """
    def Calculate_Lx(Lx):
        return Lx + 10**(0.86 * np.log10(Lx) + 4.8) - L_xuv

    if (L_xuv > 1. * 10**29):
        f_guess = L_xuv
    elif (L_xuv <= 1. * 10**29) and (L_xuv > 1. * 10**26):
        f_guess = 1. * 10**25
    elif (L_xuv <= 1. * 10**26):
        f_guess = 1. * 10**22

    Lx = optimize.fsolve(Calculate_Lx, x0=f_guess)[0]
    return Lx

#def Lx_relation_Booth(t, R_star):
#    """Gives Lx slope beyond 1 Gyr based on measurements
#		(see Booth et al. 2017). R_star in terms of solar units
#	 """
#    log_Lx = 54.65 - 2.8 * np.log10(t) + 2 * np.log10(R_star)
#    return 10.**log_Lx

def calculate_Lx_sat(L_star_bol):
    """ Typical relation (from observations) to estimate the 
    saturation X-ray luminosity given a bolometric luminosity.
    """
    return 10**(-3) * (L_star_bol * const.L_sun.cgs).value


def l_high_energy(t_i, mass_star):
    """ L_HE (UV to X-ray) as in Owen & Wu (2017).
    Parameters:
    -----------
    t_i (float): time in Myr
    mass_star (float): stellar mass in solar masses
    Returns:
    --------
    L_HE (float): X+UV luminosity in erg/s
    """
    t_sat = 100. # Myr
    L_sat = 10**(-3.5)*(mass_star) * (const.L_sun.cgs).value
    a_0 = 0.5
    if t_i < t_sat:
        L_HE = L_sat
    elif t_i >= t_sat:
        L_HE = L_sat*(t_i/t_sat)**(-1-a_0)
    return L_HE # in erg/s


import numpy as np
import math
import astropy.units as u
from astropy import constants as const

from platypos.lx_evo_and_flux import lx_evo, l_xuv_all
from platypos.lx_evo_and_flux import flux_at_planet_earth, flux_at_planet
import platypos.planet_models_LoFo14 as plmoLoFo14
import platypos.planet_model_Ot20 as plmoOt20
import platypos.beta_K_functions as bk
from platypos.mass_loss_rate_function import mass_loss_rate_forward_LO14
from platypos.mass_loss_rate_function import mass_loss_rate_forward_Ot20


def mass_planet_RK4_forward_LO14(epsilon, K_on, beta_on, planet_object,
                                 initial_step_size, t_final, track_dict):
    """USED: 4th order Runge-Kutta as numerical integration  method 
    Integrate from the current time (t_start (where planet has R0 and M0) 
    into the future taking into account photoevaporative mass loss. 

    Parameters:
    -----------
    epsilon (float): evaporation efficiency
    K_on (str): set use of K parameter on or off ("on" or "off)
    beta_on (str): set use of beta parameter on or off ("on" or "off)
    planet_object: object of planet class which contains also stellar 
                   parameters and info about stellar evo track
    step_size (float): initial step_size, variable
    t_final (float): final time of simulation
    track_dict (dict): dictionary with Lx evolutionary track parameters
    
    [NOTE: the implementation of a variable step size is somewhat preliminary.
    The step size is adjusted (made smaller or bigger depending how fast or
    slow the mass/radius changes) until the final time step greater than
    t_final. This means that if the step size in the end is e.g. 10 Myr, and
    the integration is at 4999 Myr, then last time entry will be 4999+10 ->
    5009 Myr.]

    Returns:
    --------
    t_arr (array): time array to trace mass and radius evolution
    M_arr (array): mass array with mass evolution over time (mass decrease)
    R_arr (array): radius array with radius evolution over time (from
                   thermal contraction and photoevaporative mass-loss)
    Lx_arr (array): array to trace the X-ray luminosity (mainly for
                    consistency checks)
    """ 
    # define some constants
    M_EARTH = const.M_earth.cgs.value
    Myr_to_sec = 1e6*365*86400
    
    # initialize the starting values for the X-ray luminosity at t_start,
    # the starting planet parameters, as well as beta & K (at t_start);
    # convert X-ray to XUV lum. and calculate planet's high energy 
    # incident flux
    Lx0 = lx_evo(t=track_dict["t_start"], track_dict=track_dict)
    Lxuv0 = l_xuv_all(Lx0)
    Fxuv0 = flux_at_planet(Lxuv0, planet_object.distance)

    f_env_0 = f_env = planet_object.fenv
    R0 = R = planet_object.radius
    M0 = M = planet_object.mass
    rho0 = rho = plmoLoFo14.density_planet(M0, R0)
    M_env0 = M_env = M0 - planet_object.core_mass
    M_core = planet_object.core_mass
    R_core = planet_object.core_radius
    # CRITERION for when the planet has lost all atmosphere
    # for the LoFo14 planets the core mass and thus the core radius is fixed.
    # So when the planet mass gets smaller or equal to the core mass, we
    # assume only the bare rocky core is left.

    if beta_on == "yes":
        beta = beta0 = bk.beta_fct(M0, Fxuv0, R0)
    elif beta_on == "no":
        beta = beta0 = 1.
    if K_on == "yes":
        K = K0 = bk.K_fct(planet_object.distance, M0,
                          planet_object.mass_star, R0)
    elif K_on == "no":
        K = K0 = 1.

    # since the step size is adaptive, I use lists to keep track of
    # the time, mass, radius and Lx evolution
    M_arr = []
    M_arr.append(M0)
    R_arr = []
    R_arr.append(R0) 
    t_arr = []
    t0 = t = track_dict["t_start"]
    t_arr.append(t0) 
    Lx_arr = []
    Lx_arr.append(Lx0)

    dt = initial_step_size
    # NOTE: minimum and maximum step size are HARDCODED for now (see further 
    # down in code for more details)
    min_step_size, max_step_size = 1e-2,  10.

    i = 1  # counter to track how many traced RK iterations have been performed
    j = 1  # counter to track how many RK iterations have been attempted.
    envelope_left = True  # variable to flag a planet if envelope is gone
    close_to_evaporation = False  # variable to flag if planet is close to
                                  # complete atmospheric removal

    # make list with all step sizes, even those which resulted in too drastic 
    # radius changes -> then I can check if the code gets stuck in an infinite
    # loop between make_bigger, make_smaller, make_bigger, etc..
    step_size_list = []
    
    while t <= t_final:
        #print(i, j, " t= ", t, dt)

        # This step (Lx(t) calculation) is just for me to trace Lx and check 
        # if it is correct. It is NOT required since the Lx(t) calculation is 
        # embedded in the mass_loss_rate_fancy function)
        Lx_i = lx_evo(t=t, track_dict=track_dict)

        # IMPORTANT points on the time step:
        # When the initial time step is too large OR the planet mass becomes
        # very close to the core mass (after several time steps), it can happen
        # that one of the RK substeps leads to such a large mass lost that the
        # new planet mass is smaller than the core mass.
        # Distinguish between two cases:
        # 1) initial time step is too large such that M_lost = nan after the
        # first iteration (i.e. Rk substep mass < core mass)
        # -> immediately switch to lowest possible step size and let code run
        # from there (i.e. code will make step size bigger again if necessary)
        # 2) at the end of planet evolution when the planet mass gets very
        # close to the core mass, at some point the mass lost is larger than
        # the renmaining atmosphere mass (either at the end of a coplete RK
        # iteration, or this already happends in one of the RK substeps, in
        # which case the mass lost after the complete RK step evaluates to nan
        # and no new planet radius can be calculated). In both cases the planet
        # is assumed to be fully evaporated at t_i + dt.

        # add the current step size
        step_size_list.append(dt)
        # take the last 20 entries in step_size_array and check for 
        # constant back-and-forth between two step sizes
        # e.g. [0.1, 1.0, 0.1, 1.0, ...]
        # check if difference is the same, but values in array are not all 
        # the same; also need to check if every 2nd element is the same, 
        # same goes for 2(n+1)th
        step_size_array = np.array(step_size_list)
        step_size_difference = abs(np.diff(step_size_array[-20:]))
        if (len(step_size_array) >= 20): # check only after a 20 iterations
            if (np.all(step_size_difference == step_size_difference[0]) and 
                    ~np.all(step_size_array == step_size_array[0]) and 
                    np.all(step_size_array[::2] == step_size_array[0]) and 
                    np.all(step_size_array[1::2] == step_size_array[1])):
                print("no convergence, set min. step size.")
                # if no convergence, switch to minumum step size
                dt = min_step_size
        
        # else, all is good, continue with current step size
        while (envelope_left == True):
            # go through RK iterations as long as there is envelope left
            
            # apply Runge Kutta 4th order to find next value of M_dot
            # NOTE: the mass lost in one timestep is in Earth masses
            Mdot1 = mass_loss_rate_forward_LO14(t, epsilon, K_on,
                                                beta_on, planet_object,
                                                f_env, track_dict)
            k1 = (dt * Myr_to_sec * Mdot1) / M_EARTH 
            M_05k1 = M + 0.5 * k1 # mass after 1st RK step
            M_env_05k1 = M_05k1 - M_core
            f_env_05k1 = (M_env_05k1 / M_05k1) * 100 # new envelope mass frac.
            if (i == 1) and (j == 1) and (M_05k1 < M_core):
                # then I am still in the first RK iteration, and the initial
                # step size was likely too large -> set step size to minumum
                dt = min_step_size
                j += 1
                break
            
            Mdot2 = mass_loss_rate_forward_LO14(t+0.5*dt, epsilon, K_on,
                                                beta_on, planet_object,
                                                f_env_05k1, track_dict)
            k2 = (dt * Myr_to_sec * Mdot2) / M_EARTH
            M_05k2 = M + 0.5 * k2
            M_env_05k2 = M_05k2 - M_core
            f_env_05k2 = (M_env_05k2 / M_05k2) * 100
            if (i == 1) and (j == 1) and (M_05k2 < M_core):
                dt = min_step_size
                j += 1
                break

            Mdot3 = mass_loss_rate_forward_LO14(t + 0.5*dt, epsilon, K_on,
                                                beta_on, planet_object,
                                                f_env_05k2, track_dict)
            k3 = (dt * Myr_to_sec * Mdot3) / M_EARTH
            M_k3 = M + k3
            M_env_k3 = M_k3 - M_core
            f_env_k3 = (M_env_k3 / M_k3) * 100
            if (i == 1) and (j == 1) and (M_k3 < M_core):
                dt = min_step_size
                j += 1
                break

            Mdot4 = mass_loss_rate_forward_LO14(t + dt, epsilon, K_on,
                                                beta_on, planet_object,
                                                f_env_k3, track_dict)
            k4 = (dt * Myr_to_sec * Mdot4) / M_EARTH
            # total mass lost after time-step dt
            M_lost = (k1 + 2*k2 + 2*k3 + k4) / 6. 

            # update next value of the planet mass
            M_new = M + M_lost
            M_env_new = M_new - M_core

            # now it is time to check if atmosphere is gone or 
            # if planet is close to complete atmosphere removal
            if ((np.isnan(M_lost) == True) \
                    or (np.iscomplex(M_new) == True)) \
                    and (dt == min_step_size):
                # if M_lost = nan (i.e. planet evaporates in one of the RK
                # steps) OR the four RK steps finish and the new planet mass is
                # smaller or equal to the core mass, then the planet counts as
                # evaporated!
                # if this condition is satisfied and the step size is already
                # at a minimum, then we assume the current RK iteration would
                # remove all atmosphere and only the rocky core is left at
                # t_i+dt; this terminates the code and returns the final planet
                # properties

                #print("Atmosphere has evaportated! Only bare rocky core left!"\
                #       + " STOP this madness!")

                # since the the stop criterium is reached, we assume at t_i+1
                # the planet only consists of the bare rocky core with the
                # planet mass equal to the core mass and the planet radius
                # equal to the core radius
                t_arr.append(t_arr[-1]+dt)
                M_arr.append(M_core)
                R_arr.append(R_core)
                Lx_arr.append(lx_evo(t=t_arr[-1]+dt, track_dict=track_dict))
                envelope_left = False  # set flag for complete env. removal
                j += 1
                break
                
            elif ((np.isnan(M_lost) == True) \
                    or (np.iscomplex(M_new) == True)) \
                    and (dt > min_step_size) \
                    and (close_to_evaporation == False):
                #print("close to evaporation")
                # planet close to evaporation, but since the step size is not
                # minimum yet, we set it to its minimum value and run the RK
                # iteration again (until above stopping condition is fulfilled)
                dt = min_step_size
                close_to_evaporation = True 
                # this variable is for making sure the code does not run into
                # an infinite loop when the planet is close to evaporation.
                # Once this condition is set to True, the code continues with
                # a fixed min. step size and is no longer allowed to adjust it.
                j += 1
                break
            
            # this part is new compared to the one used in the PAPER (there we
            # used a fixed step size!)
            # if you're still in the while loop at this point, then calculate
            # new radius and check how drastic the radius change would be;
            # adjust the step size if too drastic or too little
            f_env_new = (M_env_new/M_new)*100 # in %
            R_new = plmoLoFo14.calculate_planet_radius(
                                            M_core, f_env_new, t,
                                            flux_at_planet_earth(
                                                planet_object.Lbol,
                                                planet_object.distance),
                                            planet_object.metallicity
                                            )
            #print("R_new, f_new: ", R_new, f_env_new)
            #print(abs((R-R_new)/R)*100)
            
            # only adjust step size if planet is not close to complete evaporat.
            if (close_to_evaporation == False):
                #print("radius check")
                # then do the check on how much the radius changes
                # R(t_i) compared to R(t_i+dt);
                # if radius change is larger than 0.5%, make step size smaller
                # by factor 10 OR if radius change is smaller than 0.02%, make
                # step size bigger by factor 10, but not bigger or smaller than
                # max. or min. step size!
                # if radius change too much/little, do not write anything to
                # file, instead do RK iteration again with new step size

                R_change = abs((R-R_new)/R)*100 # radius change compared to
                                                # previous radius - in percent
                if (R_change > 0.5) \
                        and (t < track_dict["t_curr"]) \
                        and (dt > min_step_size):
                    dt = dt / 10.
                    j += 1
                    break

                #elif ((R_change < 1.) \
                #			and (R_change >=0.1)) \
                #			and (t < track_dict["t_curr"]) \
                #			and (dt > min_step_size):
                #    dt = dt / 10.
                #    break

                elif (R_change < (0.02)) \
                        and (t < track_dict["t_curr"]) \
                        and (dt < max_step_size):
                    dt = dt * 10.
                    j += 1
                    break    

                # NOTE: in principle I can adjust the code such that these 
                # hardcoded parameters are different for early planet evolution
                # where much more is happening typically, and late planet
                # evolution where almost no change is occurring anymore
                elif (R_change > 0.5) \
                        and (t >= track_dict["t_curr"]) \
                        and (dt > min_step_size):
                    dt = dt / 10.
                    j += 1
                    break

                elif (R_change < (0.02)) \
                        and (t >= track_dict["t_curr"]) \
                        and (dt < max_step_size):
                    dt = dt * 10
                    j += 1
                    break

                else: # if radius change is ok
                    # do sanity check: is new planet mass is still greater than
                    # the core mass? ->then there is still some atmosphere left
                    # in this case update params and go into next RK iteration
                    if ((M + M_lost) - M_core) > 0: 
                        M = M + M_lost  # new planet mass (M_lost is negative)
                        t = t_arr[-1] + dt  # updated time value t_i_plus_1
                        M_arr.append(M)
                        t_arr.append(t)
                        Lx_arr.append(lx_evo(t=t, track_dict=track_dict))

                        # calculate new envelope mass fraction:
                        M_env = M - M_core
                        f_env = (M_env/M)*100 # in %

                        # calculate new radius with new planet mass/envelope
                        # mass fraction & one time step later
                        R = plmoLoFo14.calculate_planet_radius(
                                                M_core, f_env, t,
                                                flux_at_planet_earth(
                                                    planet_object.Lbol,
                                                    planet_object.distance),
                                                planet_object.metallicity
                                                )
                        R_arr.append(R)
                        i += 1 # update step to i+1
                        j += 1

                    else:
                        # this should never happen
                        sys.exit('sth went wrong of you see this!')
                break
            
            elif (close_to_evaporation == True): 
                # if this condition is true, do not adjust step size
                # based on the radius change
                if ((M + M_lost) - M_core) > 0:
                    M = M + M_lost # new planet mass (M_lost is negative)
                    t = t_arr[-1] + dt #  updated time value t_i_plus_1
                    M_arr.append(M)
                    t_arr.append(t)
                    Lx_arr.append(lx_evo(t=t, track_dict=track_dict))

                    # calculate new envelope mass fraction:
                    M_env = M - M_core
                    f_env = (M_env/M)*100 # in %

                    # calculate new radius with new planet mass/envelope mass 
                    # fraction & one time step later
                    R = plmoLoFo14.calculate_planet_radius(
                                            M_core, f_env, t,
                                            flux_at_planet_earth(
                                                planet_object.Lbol,
                                                planet_object.distance),
                                            planet_object.metallicity
                                            )
                    R_arr.append(R)
                    i += 1 # update step to i+1
                    j += 1

                else:
                    sys.exit('sth went wrong of you see this!')
            break

        if (envelope_left == False):
            # planet has evaporated, so return last planet params for bare core
            return np.array(t_arr), np.array(M_arr), \
                   np.array(R_arr), np.array(Lx_arr)

    #print("Done!")
    return np.array(t_arr), np.array(M_arr), \
           np.array(R_arr), np.array(Lx_arr)


def mass_planet_RK4_forward_LO14_PAPER(epsilon, K_on, beta_on, 
                                       planet_object, initial_step_size,
                                       t_final, track_dict):
    """USED: 4th order Runge-Kutta as numerical integration  method 
    Integrate from the current time (t_start (where planet has R0 and M0) 
    into the future taking into account photoevaporative mass loss. 
    Step size is fixed!

    Parameters:
    -----------
    epsilon (float): evaporation efficiency
    K_on (str): set use of K parameter on or off ("on" or "off)
    beta_on (str): set use of beta parameter on or off ("on" or "off)
    planet_object: object of planet class which contains also stellar 
                   parameters and info about stellar evo track
    step_size (float): fixed
    t_final (float): final time of simulation
    track_dict (dict): dictionary with Lx evolutionary track parameters

    Returns:
    --------
    t_arr (array): time array to trace mass and radius evolution
    M_arr (array): mass array with mass evolution over time (mass decrease)
    R_arr (array): radius array with radius evolution over time (from
                   thermal contraction and photoevaporative mass-loss)
    Lx_arr (array): array to trace the X-ray luminosity (mainly for
                    consistency checks)
    """ 
    
    M_EARTH = const.M_earth.cgs.value
    Myr_to_sec = 1e6*365*86400
    
    # initialize the starting values for Lxuv(t_start), mass, density, beta, K
    Lx0 = lx_evo(t=track_dict["t_start"], track_dict=track_dict)
    Lxuv0 = l_xuv_all(Lx0)
    Fxuv0 = flux_at_planet(Lxuv0, planet_object.distance)
    
    # initial planet parameters at t_start
    f_env_0 = f_env = planet_object.fenv
    R0 = R = planet_object.radius
    M0 = M = planet_object.mass
    rho0 = rho = plmoLoFo14.density_planet(M0, R0)  # initial mean density
    M_env0 = M_env = M0 - planet_object.core_mass  # initial envelope mass
    M_core = planet_object.core_mass

    # specify beta0 and K0
    if beta_on == "yes":
        beta = beta0 = bk.beta_fct(M0, Fxuv0, R0)
    elif beta_on == "no":
        beta = beta0 = 1.
    if K_on == "yes":
        K = K0 = bk.K_fct(planet_object.distance, M0,
        				  planet_object.mass_star, R0)
    elif K_on == "no":
        K = K0 = 1.
    
    t_start = track_dict["t_start"]
    t_max = t_final
    step_size = initial_step_size
    
    # create time array for integration (with user-specified step size)
    number = math.ceil((t_max-t_start)/step_size)
    times, step_size2 = np.linspace(t_start, t_max, number,
    								endpoint=True, retstep=True)
    dt = step_size2

    # here I make lists of all the values I would like to 
    # track & output in the end:
    M_arr = []
    M_arr.append(M0)
    R_arr = []
    R_arr.append(R0)
    t_arr = []
    t_arr.append(t_start)
    Lx_arr = []
    Lx_arr.append(Lx0)
    
    # CRITERION when to stop the mass-loss
    # the LofO14 planets have a FIXED core mass and thus core radius 
    # (bare rocky core)
    R_core = planet_object.core_radius # stop when this radius is reached!

    for i in range(0, len(times)-1):   
        #print(t_arr[i])

        # this is just for me to return the Lx(t) evolution to check if 
        # it is correct (not required since the Lx(t)
        # calculation is embedded in the mass_loss_rate_fancy function)
        Lx_i = lx_evo(t=t_arr[i], track_dict=track_dict)

        Mdot1 = mass_loss_rate_forward_LO14(times[i], epsilon,
        									K_on, beta_on, planet_object,
        									f_env, track_dict)
        k1 = (dt * Myr_to_sec * Mdot1) / M_EARTH # mass lost in one timestep 
        										 # in earth masses
        M_05k1 = M + 0.5 * k1
        M_env_05k1 = M_05k1 - M_core
        f_env_05k1 = (M_env_05k1 / M_05k1) * 100

        Mdot2 = mass_loss_rate_forward_LO14(times[i]+0.5*dt, epsilon,
        									K_on, beta_on, planet_object,
        									f_env_05k1, track_dict)
        k2 = (dt * Myr_to_sec * Mdot2) / M_EARTH
        M_05k2 = M + 0.5 * k2
        M_env_05k2 = M_05k2 - M_core
        f_env_05k2 = (M_env_05k2 / M_05k2) * 100

        Mdot3 = mass_loss_rate_forward_LO14(times[i]+0.5*dt, epsilon,
        									K_on, beta_on, planet_object,
        									f_env_05k2, track_dict)
        k3 = (dt * Myr_to_sec * Mdot3) / M_EARTH
        M_k3 = M + k3
        M_env_k3 = M_k3 - M_core
        f_env_k3 = (M_env_k3 / M_k3) * 100

        Mdot4 = mass_loss_rate_forward_LO14(times[i]+dt, epsilon,
        									K_on, beta_on, planet_object,
        									f_env_k3, track_dict)
        k4 = (dt * Myr_to_sec * Mdot4) / M_EARTH
        M_lost = (k1 + 2*k2 + 2*k3 + k4)/6. # after time-step dt

        # check if planet with new mass does still have some atmosphere
        if ((M + M_lost) - M_core) > 0:
            # then planet still has some atmosphere left -> continue
            M = M + M_lost # new planet mass
            M_env = M - M_core # new envelope mass
            t = t_arr[i] + dt  # t_i_plus_1 - update time value
            f_env = (M_env/M)*100 # in %
            # calculate new radius with new planet mass/envelope mass 
            # fraction & one time step later          
            R = plmoLoFo14.calculate_planet_radius(
            								M_core, f_env, t,
            								flux_at_planet_earth(
            									planet_object.Lbol,
                                                planet_object.distance),
            								planet_object.metallicity)
            t_arr.append(t)
            M_arr.append(M)
            R_arr.append(R)
            Lx_arr.append(lx_evo(t=t, track_dict=track_dict))

        else:
            # all atmosphere is gone -> terminate
            #print("Atmosphere has evaportated! Only bare rocky" \
            # 	  + "core left! STOP this madness!")

            # if the stop criterium is reached, I add the core mass 
            # and core radius to the array at t_i+1
            t_arr.append(t_arr[-1]+dt)
            M_arr.append(M_core)
            R_arr.append(R_core)
            Lx_arr.append(lx_evo(t=t_arr[-1]+dt, track_dict=track_dict))
            return np.array(t_arr), np.array(M_arr), \
                   np.array(R_arr), np.array(Lx_arr)

    #print("Done!")
    return np.array(t_arr), np.array(M_arr), \
           np.array(R_arr), np.array(Lx_arr)


def mass_planet_RK4_forward_Ot20(epsilon, K_on, beta_on, planet_object,
								initial_step_size, t_final, track_dict):
    """USED: 4th order Runge-Kutta as numerical integration  method 
    Integrate from the current time (t_start (where planet has R0 and M0)
    into the future taking into account photoevaporative mass loss. 
    
    Parameters:
    -----------
    epsilon (float): evaporation efficiency
    K_on (str): set use of K parameter on or off ("on" or "off)
    beta_on (str): set use of beta parameter on or off ("on" or "off)
    planet_object: object of planet class which contains also stellar 
                   parameters and info about stellar evo track
    step_size (float): initial step_size, variable
    t_final (float): final time of simulation
    track_dict (dict): dictionary with Lx evolutionary track parameters
    
    [NOTE: the implementation of a variable step size is somewhat preliminary.
    The step size is adjusted (made smaller or bigger depending how fast or
    slow the mass/radius changes) until the final time step greater than
    t_final. This means that if the step size in the end is e.g. 10 Myr, and
    the integration is at 4999 Myr, then last time entry will be 4999+10 ->
    5009 Myr.]

    Returns:
    --------
    t_arr (array): time array to trace mass and radius evolution
    M_arr (array): mass array with mass evolution over time (mass decrease)
    R_arr (array): radius array with radius evolution over time (from
                   thermal contraction and photoevaporative mass-loss)
    Lx_arr (array): array to trace the X-ray luminosity (mainly for
                    consistency checks)
    """ 
    
    M_EARTH = const.M_earth.cgs.value
    Myr_to_sec = 1e6*365*86400
    
    # initialize the starting values for Lxuv(t_start), mass, density, beta, K
    Lx0 = lx_evo(t=track_dict["t_start"], track_dict=track_dict)
    Lxuv0 = l_xuv_all(Lx0)
    Fxuv0 = flux_at_planet(Lxuv0, planet_object.distance)

    # "make" initial planet at t_start
    R0 = R = planet_object.radius
    M0 = M = planet_object.mass
    rho0 = rho = plmoOt20.density_planet(M0, R0)  # initial approx. density
 
    # specify beta0 and K0
    if beta_on == "yes":
        beta = beta0 = bk.beta_fct(M0, Fxuv0, R0)
    elif beta_on == "no":
        beta = beta0 = 1.
    if K_on == "yes":
        K = K0 = bk.K_fct(planet_object.distance, M0,
        				  planet_object.mass_star, R0)
    elif K_on == "no":
        K = K0 = 1.

    M_arr = []
    M_arr.append(M0)
    R_arr = []
    R_arr.append(R0) 
    t_arr = []
    t0 = t = track_dict["t_start"]
    t_arr.append(t0) 
    Lx_arr = []
    Lx_arr.append(Lx0)

    # CRITERION when to stop the mass-loss
    # stop when this hardcoded radius is reached! 
    # (this is the minimum radius for which the volatile regime is valid)
    R_core = 2.15
    M_core = plmoOt20.calculate_mass_planet_Ot20(R_core)

    dt = initial_step_size
    # NOTE: minimum and maximum step size are HARDCODED for now (see further 
    # down in code for more details)
    min_step_size, max_step_size = 1e-2,  10.

    i = 1  # counter to track how many traced RK iterations have been performed
    j = 1  # counter to track how many RK iterations have been attempted.
    envelope_left = True  # variable to flag a planet if envelope is gone
    close_to_evaporation = False  # variable to flag if planet is close to
                                  # complete atmospheric removal

    # make list with all step sizes, even those which resulted in too drastic 
    # radius changes -> then I can check if the code gets stuck in an infinite
    # loop between make_bigger, make_smaller, make_bigger, etc..
    step_size_list = []
    
    while t <= t_final:
        #print(i, j, " t= ", t, dt)

        # This step (Lx(t) calculation) is just for me to trace Lx and check 
        # if it is correct. It is NOT required since the Lx(t) calculation is 
        # embedded in the mass_loss_rate_fancy function)
        Lx_i = lx_evo(t=t, track_dict=track_dict)

        # IMPORTANT points on the time step:
        # When the initial time step is too large OR the planet mass becomes
        # very close to the core mass (after several time steps), it can happen
        # that one of the RK substeps leads to such a large mass lost that the
        # new planet mass is smaller than the core mass.
        # Distinguish between two cases:
        # 1) initial time step is too large such that M_lost = nan after the
        # first iteration (i.e. Rk substep mass < core mass)
        # -> immediately switch to lowest possible step size and let code run
        # from there (i.e. code will make step size bigger again if necessary)
        # 2) at the end of planet evolution when the planet mass gets very
        # close to the core mass, at some point the mass lost is larger than
        # the renmaining atmosphere mass (either at the end of a coplete RK
        # iteration, or this already happends in one of the RK substeps, in
        # which case the mass lost after the complete RK step evaluates to nan
        # and no new planet radius can be calculated). In both cases the planet
        # is assumed to be fully evaporated at t_i + dt.

        # add the current step size
        step_size_list.append(dt)
        # take the last 20 entries in step_size_array and check for 
        # constant back-and-forth between two step sizes
        # e.g. [0.1, 1.0, 0.1, 1.0, ...]
        # check if difference is the same, but values in array are not all 
        # the same; also need to check if every 2nd element is the same, 
        # same goes for 2(n+1)th
        step_size_array = np.array(step_size_list)
        step_size_difference = abs(np.diff(step_size_array[-20:]))
        if (len(step_size_array) >= 20): # check only after a 20 iterations
            if (np.all(step_size_difference == step_size_difference[0]) and 
                    ~np.all(step_size_array == step_size_array[0]) and 
                    np.all(step_size_array[::2] == step_size_array[0]) and 
                    np.all(step_size_array[1::2] == step_size_array[1])):
                print("no convergence, set min. step size.")
                # if no convergence, switch to minumum step size
                dt = min_step_size
        
        # else, all is good, continue with current step size
        while (envelope_left == True):
            # go through RK iterations as long as there is envelope left
            
            # apply Runge Kutta 4th order to find next value of M_dot
            # NOTE: the mass lost in one timestep is in Earth masses
            Mdot1 = mass_loss_rate_forward_Ot20(t, epsilon,
                                                K_on, beta_on,
                                                planet_object, R,
                                                track_dict)
            k1 = (dt * Myr_to_sec * Mdot1) / M_EARTH 
            M_k1 = M + 0.5 * k1 # mass after 1st RK step
            R_k1 = plmoOt20.calculate_radius_planet_Ot20(M_k1)
            if (i == 1) and (j == 1) and (M_k1 < M_core):
                # then I am still in the first RK iteration, and the initial
                # step size was likely too large -> set step size to minumum
                dt = min_step_size
                j += 1
                break
            
            Mdot2 = mass_loss_rate_forward_Ot20(t+0.5*dt, epsilon,
                                                K_on, beta_on,
                                                planet_object, R_k1,
                                                track_dict)
            k2 = (dt * Myr_to_sec * Mdot2) / M_EARTH
            M_05k2 = M + 0.5 * k2
            R_05k2 = plmoOt20.calculate_radius_planet_Ot20(M_05k2)
            if (i == 1) and (j == 1) and (M_05k2 < M_core):
                dt = min_step_size
                j += 1
                break

            Mdot3 = mass_loss_rate_forward_Ot20(t+0.5*dt, epsilon,
                                                K_on, beta_on,
                                                planet_object, R_05k2,
                                                track_dict) 
            k3 = (dt * Myr_to_sec * Mdot3) / M_EARTH
            M_05k3 = M + k3
            R_05k3 = plmoOt20.calculate_radius_planet_Ot20(M_05k3)
            if (i == 1) and (j == 1) and (M_05k3 < M_core):
                dt = min_step_size
                j += 1
                break

            Mdot4 = mass_loss_rate_forward_Ot20(t+dt, epsilon,
                                                K_on, beta_on,
                                                planet_object, R_05k3,
                                                track_dict) 
            k4 = (dt * Myr_to_sec * Mdot4) / M_EARTH
            # total mass lost after time-step dt
            M_lost = (k1 + 2*k2 + 2*k3 + k4) / 6. 

            # update next value of the planet mass
            M_new = M + M_lost

            # now it is time to check if atmosphere is gone or 
            # if planet is close to complete atmosphere removal
            if ((np.isnan(M_lost) == True) \
                    or (np.iscomplex(M_new) == True)) \
                    and (dt == min_step_size):
                # if M_lost = nan (i.e. planet evaporates in one of the RK
                # steps) OR the four RK steps finish and the new planet mass is
                # smaller or equal to the core mass, then the planet counts as
                # evaporated!
                # if this condition is satisfied and the step size is already
                # at a minimum, then we assume the current RK iteration would
                # remove all atmosphere and only the rocky core is left at
                # t_i+dt; this terminates the code and returns the final planet
                # properties

                #print("Atmosphere has evaportated! Only bare rocky core left!"\
                #       + " STOP this madness!")

                # since the the stop criterium is reached, we assume at t_i+1
                # the planet only consists of the bare rocky core with the
                # planet mass equal to the core mass and the planet radius
                # equal to the core radius
                t_arr.append(t_arr[-1]+dt)
                M_arr.append(M_core)
                R_arr.append(R_core)
                Lx_arr.append(lx_evo(t=t_arr[-1]+dt, track_dict=track_dict))
                envelope_left = False  # set flag for complete env. removal
                j += 1
                break
                
            elif ((np.isnan(M_lost) == True) \
                    or (np.iscomplex(M_new) == True)) \
                    and (dt > min_step_size) \
                    and (close_to_evaporation == False):
                #print("close to evaporation")
                # planet close to evaporation, but since the step size is not
                # minimum yet, we set it to its minimum value and run the RK
                # iteration again (until above stopping condition is fulfilled)
                dt = min_step_size
                close_to_evaporation = True 
                # this variable is for making sure the code does not run into
                # an infinite loop when the planet is close to evaporation.
                # Once this condition is set to True, the code continues with
                # a fixed min. step size and is no longer allowed to adjust it.
                j += 1
                break
            
            # this part is new compared to the one used in the PAPER (there we
            # used a fixed step size!)
            # if you're still in the while loop at this point, then calculate
            # new radius and check how drastic the radius change would be;
            # adjust the step size if too drastic or too little
            R_new = plmoOt20.calculate_radius_planet_Ot20(M_new)
            #print("R_new, f_new: ", R_new, f_env_new)
            #print(abs((R-R_new)/R)*100)
            
            # only adjust step size if planet is not close to complete evaporat.
            if (close_to_evaporation == False):
                #print("radius check")
                # then do the check on how much the radius changes
                # R(t_i) compared to R(t_i+dt);
                # if radius change is larger than 0.5%, make step size smaller
                # by factor 10 OR if radius change is smaller than 0.02%, make
                # step size bigger by factor 10, but not bigger or smaller than
                # max. or min. step size!
                # if radius change too much/little, do not write anything to
                # file, instead do RK iteration again with new step size

                R_change = abs((R-R_new)/R)*100 # radius change compared to
                                                # previous radius - in percent
                if (R_change > 1.) \
                        and (t < track_dict["t_curr"]) \
                        and (dt > min_step_size):
                    dt = dt / 10.
                    j += 1
                    break

                #elif ((R_change < 1.) \
                #           and (R_change >=0.1)) \
                #           and (t < track_dict["t_curr"]) \
                #           and (dt > min_step_size):
                #    dt = dt / 10.
                #    break

                elif (R_change < (0.01)) \
                        and (t < track_dict["t_curr"]) \
                        and (dt < max_step_size):
                    dt = dt * 10.
                    j += 1
                    break    

                # NOTE: in principle I can adjust the code such that these 
                # hardcoded parameters are different for early planet evolution
                # where much more is happening typically, and late planet
                # evolution where almost no change is occurring anymore
                elif (R_change > 1.) \
                        and (t >= track_dict["t_curr"]) \
                        and (dt > min_step_size):
                    dt = dt / 10.
                    j += 1
                    break

                elif (R_change < (0.01)) \
                        and (t >= track_dict["t_curr"]) \
                        and (dt < max_step_size):
                    dt = dt * 10
                    j += 1
                    break

                else: # if radius change is ok
                    # do sanity check: is new planet mass is still greater than
                    # the core mass? ->then there is still some atmosphere left
                    # in this case update params and go into next RK iteration
                    if ((M + M_lost) - M_core) > 0: 
                        M = M + M_lost  # new planet mass (M_lost is negative)
                        t = t_arr[-1] + dt  # updated time value t_i_plus_1
                        M_arr.append(M)
                        t_arr.append(t)
                        Lx_arr.append(lx_evo(t=t, track_dict=track_dict))

                        # calculate new envelope mass fraction:
                        M_env = M - M_core
                        f_env = (M_env/M)*100 # in %

                        # calculate new radius with new planet mass
                        # one time step later
                        R = plmoOt20.calculate_radius_planet_Ot20(M) 
                        R_arr.append(R)
                        i += 1 # update step to i+1
                        j += 1

                    else:
                        # this should never happen
                        sys.exit('sth went wrong of you see this!')
                break
            
            elif (close_to_evaporation == True): 
                # if this condition is true, do not adjust step size
                # based on the radius change
                if ((M + M_lost) - M_core) > 0:
                    M = M + M_lost # new planet mass (M_lost is negative)
                    t = t_arr[-1] + dt #  updated time value t_i_plus_1
                    M_arr.append(M)
                    t_arr.append(t)
                    Lx_arr.append(lx_evo(t=t, track_dict=track_dict))

                    # calculate new envelope mass fraction:
                    M_env = M - M_core
                    f_env = (M_env/M)*100 # in %

                    # calculate new radius with new planet mass
                    # one time step later
                    R = plmoOt20.calculate_radius_planet_Ot20(M) 
                    R_arr.append(R)
                    i += 1 # update step to i+1
                    j += 1

                else:
                    sys.exit('sth went wrong of you see this!')
            break

        if (envelope_left == False):
            # planet has evaporated, so return last planet params for bare core
            return np.array(t_arr), np.array(M_arr), \
                   np.array(R_arr), np.array(Lx_arr)

    #print("Done!")
    return np.array(t_arr), np.array(M_arr), \
           np.array(R_arr), np.array(Lx_arr)


def mass_planet_RK4_forward_Ot20_PAPER(epsilon, K_on, beta_on, 
									   planet_object, initial_step_size, 
									   t_final, track_dict):
    """USED: 4th order Runge-Kutta as numerical integration  method 
    Integrate from the current time (t_start (where planet has R0 and M0)
    into the future taking into account photoevaporative mass loss. 

    Input:
    ----------
    epsilon: evaporation efficiency
    K_on: set use of K parameter on or off 
    beta_on: set use of beta parameter on or off
    planet_object: object of planet class which contains also stellar
                   parameters and info about stellar evo track
    step_size: initial step_size, fixed
    t_final: final time of simulation
    track_dict: dictionary with Lx evolutionary track parameters
    
    Output:
    ----------
    t_arr, M_arr, R_arr, Lx_arr: array of time, mass, radius and 
    							 Lx values from t_start to t_final
    """ 
    
    M_EARTH = const.M_earth.cgs.value
    Myr_to_sec = 1e6*365*86400
    
    # initialize the starting values for Lxuv(t_start), mass, density, beta, K
    Lx0 = lx_evo(t=track_dict["t_start"], track_dict=track_dict)
    Lxuv0 = l_xuv_all(Lx0)
    Fxuv0 = flux_at_planet(Lxuv0, planet_object.distance)
    
    # "make" initial planet at t_start
    R0 = R = planet_object.radius
    M0 = M = planet_object.mass
    rho0 = rho = plmoOt20.density_planet(M0, R0)  # initial approx. density
    
    # specify beta0 and K0
    if beta_on == "yes":
        beta = beta0 = bk.beta_fct(M0, Fxuv0, R0)
    elif beta_on == "no":
        beta = beta0 = 1.
    if K_on == "yes":
        K = K0 = bk.K_fct(planet_object.distance, M0,
        				  planet_object.mass_star, R0)
    elif K_on == "no":
        K = K0 = 1.

    # create time array for integration (with user-specified step size)
    t_start, t_max = track_dict["t_start"], t_final
    step_size = initial_step_size
    number = math.ceil((t_max-t_start)/step_size)
    times, step_size2 = np.linspace(t_start, t_max, number,
    								endpoint=True, retstep=True)
    dt = step_size2

    # make lists of all the values wwe want to track & output in the end:
    M_arr = []
    M_arr.append(M0)
    R_arr = []
    R_arr.append(R0)
    t_arr = []
    t_arr.append(t_start)
    Lx_arr = []
    Lx_arr.append(Lx0)
    
    # CRITERION when to stop the mass-loss
    R_core = 2.15 # stop when this radius is reached! (this is the minimum
                  # radius for which the volatile regime is valid)
    M_core = plmoOt20.calculate_mass_planet_Ot20(R_core)

    for i in range(0, len(times)-1):   
        # this is just for me to return the Lx(t) evolution to check if it is
        # correct (not required since the Lx(t) calculation is embedded in 
        # the mass_loss_rate_fancy function)
        Lx_i = lx_evo(t=t_arr[i], track_dict=track_dict)

        Mdot1 = mass_loss_rate_forward_Ot20(times[i], epsilon, 
        									K_on, beta_on, 
        									planet_object, R, 
        									track_dict)
        k1 = (dt * Myr_to_sec * Mdot1) / M_EARTH # in earth masses
        M_05k1 = M + 0.5 * k1     
        R_05k1 = plmoOt20.calculate_radius_planet_Ot20(M_05k1)
        
        Mdot2 = mass_loss_rate_forward_Ot20(times[i]+0.5*dt, epsilon,
        									K_on, beta_on,
        									planet_object, R_05k1,
        									track_dict)
        k2 = (dt * Myr_to_sec * Mdot2) / M_EARTH
        M_05k2 = M + 0.5 * k2
        R_05k2 = plmoOt20.calculate_radius_planet_Ot20(M_05k2)
        
        Mdot3 = mass_loss_rate_forward_Ot20(times[i]+0.5*dt, epsilon,
        									K_on, beta_on,
        									planet_object, R_05k2,
        									track_dict) 
        k3 = (dt * Myr_to_sec * Mdot3) / M_EARTH
        M_05k3 = M + 0.5 * k3
        R_05k3 = plmoOt20.calculate_radius_planet_Ot20(M_05k3)
        
        Mdot4 = mass_loss_rate_forward_Ot20(times[i]+dt, epsilon,
        									K_on, beta_on,
        									planet_object, R_05k3,
        									track_dict) 
        k4 = (dt * Myr_to_sec * Mdot4) / M_EARTH
        M_lost = (k1 + 2*k2 + 2*k3 + k4)/6. # mass lost after time-step dt

        # check if planet with new mass does still have some atmosphere
        if ((M + M_lost) - M_core) > 0:
            # then planet still has some atmosphere left -> continue
            M = M + M_lost # new planet mass (M_lost is negative)
            # t_i_plus_1 - update time value
            t = t_arr[-1] + dt
            # calculate new radius with new planet mass
            R = plmoOt20.calculate_radius_planet_Ot20(M)
            M_arr.append(M)
            R_arr.append(R)
            t_arr.append(t)
            Lx_arr.append(lx_evo(t=t, track_dict=track_dict))
            
        else:
            # all atmosphere is gone (based on criterion set at the top)
            #print("Atmosphere has evaportated! Only bare rocky core" \
            #	 + "left! STOP this madness!")

            # if the stop criterium is reached, I add the core mass 
            # and core radius to the array at t_i+1
            
            t_arr.append(t_arr[-1]+dt)
            M_arr.append(M_core)
            R_arr.append(R_core)
            Lx_arr.append(lx_evo(t=t_arr[-1]+dt, track_dict=track_dict))
            return np.array(t_arr), np.array(M_arr), \
                   np.array(R_arr), np.array(Lx_arr)
    
    # if planet survives, output the final arrays
    #Lx_arr[i+1] = lx_evo(t=t_arr[-1]+dt, track_dict=track_dict)
    #Lx_arr.append(lx_evo(t=t_arr[-1]+dt, track_dict=track_dict))
    #print("Done!")
    return np.array(t_arr), np.array(M_arr), \
           np.array(R_arr), np.array(Lx_arr)



### notify_message.py ###

import os
from io import StringIO
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart

def message(subject="Python Notification", text="", img=None, attachment=None):
    # build message contents
    msg = MIMEMultipart()
    msg['Subject'] = subject  # add in the subject
    msg.attach(MIMEText(text))  # add text contents

    # check if we have anything given in the img parameter
    if img is not None:
        # if we do, we want to iterate through the images, so let's check that
        # what we have is actually a list
        if type(img) is not list:
            img = [img]  # if it isn't a list, make it one
        # now iterate through our list
        for one_img in img:
            img_data = open(one_img, 'rb').read()  # read the image binary data
            # attach the image data to MIMEMultipart using MIMEImage, we add
            # the given filename use os.basename
            msg.attach(MIMEImage(img_data, name=os.path.basename(one_img)))

    # we do the same for attachments as we did for images
    if attachment is not None:
        if type(attachment) is not list:
            attachment = [attachment]  # if it isn't a list, make it one
        with open(attachment, 'rb') as f:
            # read in the attachment using MIMEApplication
            file = MIMEApplication(
                f.read(),
                name=os.path.basename(attachment)
            )
        # here we edit the attached file metadata
        file['Content-Disposition'] = f'attachment; filename="{os.path.basename(attachment)}"'
        msg.attach(file)  # finally, add the attachment to our message object
    return msg


### notify_smtp ###

import smtplib
import socket
from io import StringIO

def send(msg, email, pwd, server='smtp.gmail.com', port=587):
    # contain following in try-except in case of momentary network errors
    try:
        # initialise connection to email server, the default is Outlook
        smtp = smtplib.SMTP(server, port)
        # this is the 'Extended Hello' command, essentially greeting our SMTP or ESMTP server
        smtp.ehlo()
        # this is the 'Start Transport Layer Security' command, tells the server we will
        # be communicating with TLS encryption
        smtp.starttls()
        
        # read email and password from file
#         with open(path_to_username_password_files+'/email.txt', 'r') as fp:
#             email = fp.read()
#         with open(path_to_username_password_files+'/password.txt', 'r') as fp:
#             pwd = fp.read()
            
        # login to the email server
        smtp.login(email, pwd)
        # send notification to self
        smtp.sendmail(email, email, msg.as_string())
        # disconnect from the server
        smtp.quit()
    except socket.gaierror:
        print("Network connection error, email not sent.")


import setuptools

# use readme as long description
with open("README.md", "r") as fh:
    long_description = fh.read()

# read in required packages from requirements.txt
with open("requirements.txt", "r") as fh:
    requirements = fh.read().splitlines()

#get version number
from subprocess import Popen, PIPE

try:
    process = Popen(['git', 'describe', '--exact-match', '--tags'], stdout=PIPE, stderr=PIPE)
    stdout, stderr = process.communicate()
    version = stdout.decode('ascii')
except:
    version = '0.0.0'

# set up setuptools
setuptools.setup(
    name="platypos",
    version="1.1.16",
    author='Laura Ketzer',
    author_email='lketzer@aip.de',
    description='PLAneTarY PhOtoevaporation Simulator',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lketzer/platypos",
    install_requires=requirements,
    keywords='astronomy, exoplanets, ...',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        'Development Status :: 3 - Alpha', # 2 - Pre-Alpha, # 3 - Alpha, 4 - Beta, 5 - Production/Stable, 6 - Mature, 7 - Inactive (1 - Planning)
        'Environment :: Console',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)




import os
import numpy as np
import pandas as pd
from scipy.optimize import fsolve
import scipy.optimize as optimize
import astropy.units as u
from astropy import constants as const

from platypos.lx_evo_and_flux import flux_at_planet_earth, flux_at_planet
from platypos.mass_evolution_function import mass_planet_RK4_forward_LO14


class Planet_LoFo14():
    """
    Need star and planet dictionary to initialize a LoFo planet object.

    Structure of star_dictionary: {'star_id': "dummySun", 'mass': mass_star,
                                   'radius': radius_star, 'age': age_star,
                                   'L_bol': L_bol, 'Lx_age': Lx_age}
    Structure of planet_dict: {"core_mass": m_c, "fenv": f,
                              "distance": a, "metallicity": metal}

    NOTE: There are actually three options for planet_dict:
         Case 1) An artificial planet with given core mass and envelope mass
                 fraction (M_core & f_env)
                -> in this case we need to calculate the mass and radius of
                the planet (M_pl & R_pl)

         Case 2) An observed planet with a known mass (we have M_pl & R_pl
                 & f_env) -> M_core

         Case 3) An observed planet with radius and mass measurement, plus
                 core mass is specified
                 -> need to calculate/estimate envelope mass fraction

    Additional Input: Lx_sat_info (only needed if you want to scale the
                      1 & 5 Gyr X-ray luminosities for non-solar-mass stars
    """

    def __init__(self, star_dictionary, planet_dict, Lx_sat_info=None):
        '''Initialize a new instance (=object) of Planet_LoFo14

        Parameters:
        -----------


        '''

        # initialize stellar parameters
        self.star_id = star_dictionary['star_id']
        self.mass_star = star_dictionary['mass']
        self.radius_star = star_dictionary['radius']
        self.age = star_dictionary['age']
        self.Lbol = star_dictionary['L_bol'] * const.L_sun.cgs.value
        self.Lx_age = star_dictionary['Lx_age']
        self.Lx_sat_info = Lx_sat_info

        # initialize planet parameters based on the input dictionary
        # the following 3 parameters are the same for all planets!
        self.distance = planet_dict["distance"]
        self.metallicity = planet_dict["metallicity"]  # solarZ or enhZ
        self.flux = flux_at_planet_earth(self.Lbol, self.distance)

        # flag which tracks if planet has been evolved and result file exists
        self.has_evolved = False
        self.planet_id = "dummy" # set planet name later with set_name
        self.Lx_sat_info = Lx_sat_info

        # the remaining parameters depend on the input dictionary (Case 1, 2, 3)
        while True:
            try:
                # check if envelope mass fraction is specified, then Case 1
                planet_dict["fenv"]
                self.planet_info = "Case 1 - artificial planet"\
                                    + " (with M_core & f_env)"
                # Case 1: artificial planet with fenv & M_core given, need to
                #         calculate the total mass and radius
                self.fenv = planet_dict["fenv"]
                self.core_mass = planet_dict["core_mass"]
                self.calculate_core_radius()
                self.calculate_planet_mass()
                self.calculate_planet_radius()
                break
            except KeyError:  # if no f_env is provided, then we are dealing
                              # with Case 2 or 3
                while True:
                    try:
                        # check if planet mass is provided, then Case 3
                        planet_dict["mass"]
                        # Case 3: An observed planet with radius and mass
                        #         measurement, plus core mass is specified;
                        #         need to calculate/estimate envelope mass frac.
                        self.planet_info = "Case 3 - obs. planet with radius"\
                        					+ " & mass measurement (and core "\
                        					+ "mass estimate)"
                        self.core_mass = planet_dict["core_mass"]
                        self.mass = planet_dict["mass"]
                        self.radius = planet_dict["radius"]
                        self.calculate_core_radius()
                        self.solve_for_fenv()  # get for envelope mass fraction

                        # Note to self: add sanity check to make sure the mass
                        # with the calculated fenv matches the input mass!
                        # Note to self: add sanity check to make sure the radius
                        # with the calculated fenv matches the input radius!
                        break
                    except KeyError:
                        # no mass and fenv given -> Case 2
                        self.planet_info = "Case 2 - obs. planet with"\
                        				    + " radius, but no mass measurement"
                        # Case 2 - observed planet with a without a mass, but
                        # core mass estimate, need to calculate fenv
                        self.core_mass = planet_dict["core_mass"]
                        self.radius = planet_dict["radius"]
                        self.calculate_core_radius()
                        self.solve_for_fenv()  # get envelope mass fraction
                        self.calculate_planet_mass()
                        # Note to self: add sanity check to make sure the radius
                        # with the calculated fenv matches the input radius!
                        break
                break

    # Class Methods
    def calculate_planet_mass(self):
        """ Planet mass determined by core and atmosphere mass
            (specified in terms of envelope mass fraction [in %]). """
        self.mass = self.core_mass/(1-(self.fenv/100))

    def calculate_planet_core_mass(self):
        """ Core mass determined by planet mass and envelope mass
            (specified in terms of envelope mass fraction [%]). """
        self.core_mass = self.mass*(1-(self.fenv/100))

    def calculate_core_radius(self):
        """M-R relation for rock/iron Earth-like core. (no envelope)"""
        self.core_radius = (self.core_mass**0.25)

    def solve_for_fenv(self):
        """ For known core and planet radius, core mass, age and flux,
            solve for envelope mass fraction."""
        if self.radius == self.core_radius:
            self.fenv = 0.0
        else:
            def calculate_fenv(fenv):
                age_exponent = {"solarZ": -0.11, "enhZ": -0.18}
                return -self.radius + self.core_radius + (2.06 \
                       * (self.core_mass/(1 - (fenv / 100)))**(-0.21) \
                       * (fenv / 5)**0.59 * (self.flux)**0.044 \
                       * ((self.age / 1e3) \
                            / 5)**(age_exponent[self.metallicity]))
            f_guess = 0.1
            fenv = optimize.fsolve(calculate_fenv, x0=f_guess)[0]
            self.fenv = fenv
            # if fenv1 != fenv:
            #    print("Sth went wrong in solving for\
            #          the envelope mass fraction! Check!")


    def calculate_R_env(self):
        """ Check Planet_models_LoFo14.py for details on input and
            output parameters;
            R_env ~ t**0.18 for *enhanced opacities*;
            R_env ~ t**0.11 for *solar metallicities*
        """
        age_exponent = {"solarZ": -0.11, "enhZ": -0.18}
        R_env = 2.06 * self.mass**(-0.21) * (self.fenv / 5)**0.59 * \
            self.flux**0.044 * \
            ((self.age / 1e3) / 5)**(age_exponent[self.metallicity])
        return R_env  # in units of R_earth


    def calculate_planet_radius(self):
        """ Check Planet_models_LoFo14.py for details"""
        self.radius = self.core_radius + self.calculate_R_env()


    def set_name(self, t_final, initial_step_size,
                 epsilon, K_on, beta_on, evo_track_dict):
        """ OBSOLETE
        Function to set the right planet name based on the track specified.
        This can then be used to check if a particular planet on a particular
        track has already evolved (i.e. if outpufile exists). """
        self.planet_id = "planet_" \
                         + str(np.round(self.distance, 3)).replace('.', 'dot') \
                         + '_Mcore' \
                         + str(np.round(self.core_mass, 3)).replace(".", "dot" )\
                         + '_fenv' \
                         + str(np.round(self.fenv, 3)) + '_' + self.metallicity \
                         + '_Mstar' \
                         + str(np.round(self.mass_star, 3)).replace(".", "dot") \
                         + "_K_" + K_on + "_beta_" + beta_on \
                         + "_track_" + str(evo_track_dict["t_start"]) + "_" \
                         + str(evo_track_dict["t_sat"]) + "_" + str(t_final) \
                         + "_" + str(evo_track_dict["Lx_max"]) + "_" \
                         + str(evo_track_dict["dt_drop"]) + "_" \
                         + str(evo_track_dict["Lx_drop_factor"])


    def generate_planet_id(self, t_final, 
                           planet_folder_id, evo_track_dict):
        """ Similar as set_name; produces a planet_id which can be used 
            for saving the results. """
        self.planet_id = planet_folder_id + "_track_" \
                         + str(evo_track_dict["t_start"]) + "_" \
                         + str(evo_track_dict["t_sat"]) + "_" \
                         + str(t_final) + "_" + str(evo_track_dict["Lx_max"]) \
                         + "_" + str(evo_track_dict["dt_drop"]) + "_" \
                         + str(evo_track_dict["Lx_drop_factor"])


    def write_general_params_to_file(self, path_for_saving, 
                                     planet_folder_id, evo_track_dict):
        """produces various files:
            - a file with the initial planet parameters (name is the
              planet folder name, e.g. planet_001.txt), 
            - a file with the host star parameters (host_star.txt), 
            - a file with the corresponding track parameters starting with 
              "track_params_"
        """ 
        # create a file (e.g. planet_XXXX.txt) which contains the
        # initial planet params
        if not os.path.exists(path_for_saving + planet_folder_id + ".txt"):
            with open(path_for_saving + planet_folder_id + ".txt", "w") as p:
                planet_params = "a,core_mass,fenv,mass,radius,metallicity,age\n" \
                                 + str(self.distance) + "," \
                                 + str(self.core_mass) + "," \
                                 + str(self.fenv) + "," + str(self.mass) \
                                 + "," + str(self.radius) + "," \
                                 + self.metallicity + "," + str(self.age)
                p.write(planet_params)

        # create a file (track_params_planet_....txt) which contains
        # the track parameters
        if not os.path.exists(path_for_saving + "track_params_" \
                              + self.planet_id + ".txt"):
            with open(path_for_saving + "track_params_" + self.planet_id + ".txt", "w") as t:
                track_params = "t_start,t_sat,t_curr,t_5Gyr,Lx_max,Lx_curr," \
                                + "Lx_5Gyr,dt_drop,Lx_drop_factor\n" \
                                + str(evo_track_dict["t_start"]) + "," \
                                + str(evo_track_dict["t_sat"]) + ","  \
                                + str(evo_track_dict["t_curr"]) + "," \
                                + str(evo_track_dict["t_5Gyr"]) + "," \
                                + str(evo_track_dict["Lx_max"]) + "," \
                                + str(evo_track_dict["Lx_curr"]) + "," \
                                + str(evo_track_dict["Lx_5Gyr"]) + "," \
                                + str(evo_track_dict["dt_drop"]) + "," \
                                + str(evo_track_dict["Lx_drop_factor"])
                t.write(track_params)

        # create a file which contains the host star parameters
        if not os.path.exists(path_for_saving + "host_star_properties.txt"):
            with open(path_for_saving + "host_star_properties.txt", "w") as s:
                star_params = "star_id,mass_star,radius_star,age,Lbol,Lx_age\n" \
                              + self.star_id + "," + str(self.mass_star) \
                              + "," + str(self.radius_star) \
                              + "," + str(self.age) \
                              + "," + str(self.Lbol / const.L_sun.cgs.value) \
                              + "," + str(self.Lx_age)
                s.write(star_params)


    def write_final_params_to_file(self, results_df, path_for_saving):
        """ Create file with only the final time, mass and radius parameters. """ 
        
        # create another file, which contains the final parameters only
        if not os.path.exists(path_for_saving \
                              + self.planet_id + "_final.txt"):
            with open(path_for_saving + self.planet_id + "_final.txt", "w") as p:
                # get last element (final time, mass, radius)
                t_final = results_df["Time"].iloc[-1]
                R_final = results_df["Radius"].iloc[-1]
                M_final = results_df["Mass"].iloc[-1]
                f_env_final = ((M_final - self.core_mass) / M_final) * 100  # %
                planet_params = "a,core_mass,time,fenv,mass,radius,metallicity,track\n" \
                                + str(self.distance) + "," \
                                + str(self.core_mass) + "," \
                                + str(t_final) + "," \
                                + str(f_env_final) + "," \
                                + str(M_final) + "," \
                                + str(R_final) + "," \
                                + self.metallicity + "," \
                                + self.planet_id
                p.write(planet_params)

        
    def evolve_forward(self, t_final,
                       initial_step_size,
                       epsilon, K_on, beta_on,
                       evo_track_dict,
                       path_for_saving,
                       planet_folder_id):
        """ Call this function to make the planet evolve and 
        create file with mass and radius evolution.
        See Mass_evolution_function.py for details on the integration."""

        if os.path.exists(path_for_saving + self.planet_id + ".txt"):
            # planet already exists
            self.has_evolved = True
            df = pd.read_csv(path_for_saving + self.planet_id + ".txt")
        else:
            #print("Planet: ", self.planet_id+".txt")
            # call mass_planet_RK4_forward_LO14 to start the integration
            t, M, R, Lx = mass_planet_RK4_forward_LO14(
                                    epsilon=epsilon,
                                    K_on=K_on,
                                    beta_on=beta_on,
                                    planet_object=self,
                                    initial_step_size=initial_step_size,
                                    t_final=t_final,
                                    track_dict=evo_track_dict
                                    )
            
            ### TO DO: move this to mass_planet_RK4_forward_LO14 -
            # > make it return dataframe
            # add results to dataframe and save
            df = pd.DataFrame({"Time": t, "Mass": M, "Radius": R, "Lx": Lx})
            df.to_csv(path_for_saving + self.planet_id + ".txt", index=None)
            self.has_evolved = True  # set evolved-flag to True
            
        return df
    

    def evolve_forward_and_create_full_output(self, t_final,
                                              initial_step_size,
                                              epsilon,
                                              K_on,
                                              beta_on,
                                              evo_track_dict,
                                              path_for_saving,
                                              planet_folder_id
                                              ):
        """ This is the master-function which needs to be called to 
        evolve the planet and at the same time create all necessary 
        output files which contain useful data about initial params, 
        host star params, tracks etc..."""
        
        # create planet id for file names
        self.generate_planet_id(t_final, planet_folder_id, evo_track_dict)
        # create file with initial planet, host star and track params
        self.write_general_params_to_file(path_for_saving,
                                          planet_folder_id,
                                          evo_track_dict)
        
        results_df = self.evolve_forward(t_final, initial_step_size,
                                         epsilon, K_on, beta_on,
                                         evo_track_dict,
                                         path_for_saving,
                                         planet_folder_id)
        
        # create file with final planet params
        self.write_final_params_to_file(results_df, path_for_saving)

    def read_results(self, file_path):
        """  read in results file and return dataframe. """
        # planet exists, has been evolved and results can be read in
        if self.has_evolved == True:
            df = pd.read_csv(file_path+self.planet_id+".txt",
                             float_precision='round_trip')
            # 'round_trip': to avoid pandas doing sth. weird to the last digit
            # Info: Pandas uses a dedicated dec 2 bin converter that
            # a compromisesccuracy in preference to speed.
            # Passing float_precision='round_trip' to read_csv fixes this.
            #t, M, R, Lx = df["Time"].values, df["Mass"].values,\
            #              df["Radius"].values, df["Lx"].values
            return df #t, M, R, Lx
        else:
            print("Planet has not been evolved & no result file exists.")

    def evolve_backwards(self):
        raise NotImplementedError("Coming soon! :)")



import numpy as np

import astropy.units as u
from astropy import constants as const


def calculate_mass_planet_Ot20(R_pl):
    """ Estimate a planet mass (in Earth masses) based on the mass-radius
    relation derived for the volatile rich regime in Otegi et al. 2020
    (radius planet larger than 2.115 Earth radii, or density < 3.3 g/cc).
    Mass has to be smaller than 120 Earth masses. Radius should be given
    in Earth radii.
    """

    M_earth = const.M_earth.cgs.value
    R_earth = const.R_earth.cgs.value

    if (type(R_pl) == int) or (type(R_pl) == float) \
       or (type(R_pl) == np.float64):  # if R is single value
        M_p_volatile = 1.74*(R_pl)**1.58  # if rho < 3.3 g/cm^3
        rho_volatile = (M_p_volatile*M_earth) \
                        / (4/3*np.pi*(R_pl*R_earth)**3)
        if (rho_volatile >= 3.3):
            raise Exception("Planet with this radius is too small and"\
				 			+ " likely rocky; use LoFo14 models instead.")
        else:
            if (M_p_volatile >= 120):
                raise Exception("Planet too massive. M-R relation only valid"\
                                + " for <120 M_earth.")
            else:
                return M_p_volatile

    elif len(R_pl) > 1:  # if R is array
        Ms = []
        for i in range(len(R_pl)):
            M_p_volatile_i = 1.74*(R_pl[i])**1.58  # if rho < 3.3 g/cm^3
            rho_volatile_i = (M_p_volatile_i * M_earth) \
                             / (4/3*np.pi*(R_pl[i]*R_earth)** 3)
            if (rho_volatile_i >= 3.3) or (M_p_volatile_i >= 120):
                M_i = np.nan
            else:
                M_i = M_p_volatile_i
            Ms.append(M_i)
        Ms = np.array(Ms)
        return Ms


def calculate_radius_planet_Ot20(M_pl):
    """ Estimate a planet radius (in Earth radii) based on the mass-radius
    relation derived for the volatile rich regime in Otegi et al. 2020.
    Mass needs to be between ~5.7 and 120 Earth masses and should be given 
    in Earth masses.
    """

    M_earth = const.M_earth.cgs.value
    R_earth = const.R_earth.cgs.value

    if (type(M_pl) == int) or (type(M_pl) == float) \
            or (type(M_pl) == np.float64):  # if M is single value
        # R_p_volatile = 0.7*M_pl**0.63 # if rho < 3.3 g/cm^3
        R_p_volatile = (M_pl/1.74)**(1./1.58)
        rho_volatile = (M_pl*M_earth) \
                        / (4/3*np.pi*(R_p_volatile*R_earth)**3)
        if (rho_volatile >= 3.3):
            raise Exception("Planet with this mass/radius is too small and"\
                            + " likely rocky; use LoFo14 models instead.")
        else:
            if (M_pl >= 120):
                raise Exception("Planet too massive. M-R relation only"\
                                + " valid for <120 M_earth.")
            else:
                return R_p_volatile

    elif len(M_pl) > 1:  # if M is array
        Rs = []
        for i in range(len(M_pl)):
            # R_p_volatile_i = 0.7*M_pl[i]**0.63 # if rho < 3.3 g/cm^3
            R_p_volatile_i = (M_pl[i]/1.74)**(1./1.58)
            rho_volatile_i = (M_pl[i]*M_earth)\
                            / (4/3*np.pi*(R_p_volatile_i*R_earth)**3)
            if (rho_volatile_i >= 3.3) or (M_pl[i] >= 120):
                R_i = np.nan
            else:
                R_i = R_p_volatile_i
            Rs.append(R_i)
        Rs = np.array(Rs)
        return Rs


def density_planet(M_p, R_p):
    """ Calculate the mean density in cgs units using the radius and
    mass of the planet.

    Parameters:
    -----------
    M_p (float): mass of the planet (in Earth masses)
    R_p (float): radius of the planet (in Earth radii)

    Returns:
    --------
    rho (float): density in cgs units (g/ccm)
    """

    rho = (M_p*const.M_earth.cgs) \
        / (4./3*np.pi*(R_p*const.R_earth.cgs)**3)
    return rho.value



import astropy.units as u
from astropy import constants as const

from platypos.lx_evo_and_flux import lx_evo, l_xuv_all
from platypos.lx_evo_and_flux import flux_at_planet_earth, flux_at_planet
import platypos.planet_models_LoFo14 as plmoLoFo14
import platypos.planet_model_Ot20 as plmoOt20
import platypos.beta_K_functions as bk


def mass_loss_rate_forward_LO14(t_, epsilon, K_on, beta_on,
                                planet_object, f_env, track_dict):
    """ Calculate the XUV-induced mass-loss rate at any given time step
    (of the integration) for a Lopez & Fortney (2014) planet with rocky core
    and H/He envelope (see Planet_models_LoFo14 for details).
    (see e.g. Lammer et al. 2003, Owen & Wu 2013; Lopez & Fortney 2013 for
    details on XUV-induced mass loss/ photoevaporation)

    Parameters:
    -----------
    t_ (float): time (in Myr)
    epsilon (float): evaporation efficiency
    K_on (float): set use of K parameter on or off ("on" or "off)
    beta_on (float): set use of beta parameter on or off ("on" or "off)
    planet_object (class obj): object of planet class which contains
                               planetray & stellar parameters (here we read out
                               the core mass, and together with the envelope
                               mass fraction at time t_ we can calcualte the
                               current mass and radius of the planet)
    f_env (float): envelope mass fraction at time t_
    track_dict (dict): dictionary with evolutionary track parameters

    Returns:
    --------
    mass-loss rate in cgs units (g/s) -> NOTE: mass-loss rate is negative!
    """

    # calculate X-ray luminosity and planet flux at time t_
    Lx = lx_evo(t=t_, track_dict=track_dict)
    Fxuv = flux_at_planet(l_xuv_all(Lx), planet_object.distance)

    # calculate current planet density
    R_p = plmoLoFo14.calculate_planet_radius(planet_object.core_mass,
                                             f_env, t_,
                                             flux_at_planet_earth(
                                                    planet_object.Lbol,
                                                    planet_object.distance),
                                             planet_object.metallicity)
    M_p = plmoLoFo14.calculate_planet_mass(planet_object.core_mass, f_env)
    rho_p = rho = plmoLoFo14.density_planet(M_p, R_p)  # mean density

    # specify beta and K
    if beta_on == "yes":
        beta = bk.beta_fct(M_p, Fxuv, R_p)
    elif beta_on == "no":
        beta = 1.
    if K_on == "yes":
        K = bk.K_fct(planet_object.distance, M_p, 
        			 planet_object.mass_star, R_p)
    elif K_on == "no":
        K = 1.

    num = 3. * beta**2. * epsilon * Fxuv
    denom = 4. * const.G.cgs * K * rho_p
    M_dot = num / denom

    # NOTE: I define my mass loss rate to be negative, this way the
    # planet loses mass as I integrate forward in time.
    return -(M_dot.decompose(bases=u.cgs.bases)).value  # in cgs units [g/s]


def mass_loss_rate_forward_Ot20(t_, epsilon, K_on, beta_on,
                                planet_object, radius_at_t, track_dict):
    """ Calculate the XUV-induced mass-loss rate at any given time step
    (of the integration) for a planet which follows the "mature" mass-radius
    relation as given in Otegi et al. (2020) (see Planet_models_Ot20 for
    details).
    (see e.g. Lammer et al. 2003, Owen & Wu 2013; Lopez & Fortney 2013 for
    details on XUV-induced mass loss/ photoevaporation)

    Parameters:
    -----------
    t_ (float): time (in Myr)
    epsilon (float): evaporation efficiency
    K_on (float): set use of K parameter on or off ("on" or "off)
    beta_on (float): set use of beta parameter on or off ("on" or "off)
    planet_object (class obj): object of planet class which contains
                               planetray & stellar parameters
    radius_at_t (float): current planetary radius at time t_; used to estimate
                         the new planet mass at time t_
    track_dict (dict): dictionary with evolutionary track parameters

    Returns:
    --------
    mass-loss rate in cgs units (g/s) -> NOTE: mass-loss rate is negative!
    """

    # calculate X-ray luminosity and planet flux at time t_
    Lx = lx_evo(t=t_, track_dict=track_dict)
    Fxuv = flux_at_planet(l_xuv_all(Lx), planet_object.distance)

    # estimate planetary mass using the Otegi 2020 mass-radius relation
    # and then calculate current mean density
    M_p = plmoOt20.calculate_mass_planet_Ot20(radius_at_t)
    rho_p = rho = plmoOt20.density_planet(M_p, radius_at_t)  # mean density

    # specify beta and K
    if beta_on == "yes":
        beta = bk.beta_fct(M_p, Fxuv, radius_at_t)
    elif beta_on == "no":
        beta = 1.
    if K_on == "yes":
        K = bk.K_fct(planet_object.distance, M_p,
                     planet_object.mass_star, radius_at_t)
    elif K_on == "no":
        K = 1.

    num = 3. * beta**2. * epsilon * Fxuv
    denom = 4. * const.G.cgs * K * rho_p
    M_dot = num / denom

    # NOTE: I define my mass loss rate to be negative, this way the
    # planet loses mass as I integrate forward in time.
    return -(M_dot.decompose(bases=u.cgs.bases)).value  # in cgs units [g/s]
